# SPDX-License-Identifier: GPL-2.0-or-later
# FRR ELF xref extractor
#
# Copyright (C) 2020  David Lamparter for NetDEF, Inc.

import sys
import os
import struct
import re
import traceback

json_dump_args = {}

try:
    import ujson as json

    json_dump_args["escape_forward_slashes"] = False
except ImportError:
    import json

import argparse

from clippy.uidhash import uidhash
from clippy.elf import *
from clippy import frr_top_src, CmdAttr, elf_notes
from tiabwarfo import FieldApplicator
from xref2vtysh import CommandEntry

try:
    with open(os.path.join(frr_top_src, "python", "xrefstructs.json"), "r") as fd:
        xrefstructs = json.load(fd)
except FileNotFoundError:
    sys.stderr.write(
        """
The "xrefstructs.json" file (created by running tiabwarfo.py with the pahole
tool available) could not be found.  It should be included with the sources.
"""
    )
    sys.exit(1)

# constants, need to be kept in sync manually...

XREFT_EVENTSCHED = 0x100
XREFT_LOGMSG = 0x200
XREFT_DEFUN = 0x300
XREFT_INSTALL_ELEMENT = 0x301

# LOG_*
priovals = {}
prios = ["0", "1", "2", "E", "W", "N", "I", "D"]


class XrelfoJson(object):
    def dump(self):
        pass

    def check(self, wopt):
        yield from []

    def to_dict(self, refs):
        pass


class Xref(ELFDissectStruct, XrelfoJson):
    struct = "xref"
    fieldrename = {"type": "typ"}
    containers = {}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._container = None
        if self.xrefdata:
            self.xrefdata.ref_from(self, self.typ)

    def container(self):
        if self._container is None:
            if self.typ in self.containers:
                self._container = self.container_of(self.containers[self.typ], "xref")
        return self._container

    def check(self, *args, **kwargs):
        if self._container:
            yield from self._container.check(*args, **kwargs)


class Xrefdata(ELFDissectStruct):
    struct = "xrefdata"

    # uid is all zeroes in the data loaded from ELF
    fieldrename = {"uid": "_uid"}

    def ref_from(self, xref, typ):
        self.xref = xref

    @property
    def uid(self):
        if self.hashstr is None:
            return None
        return uidhash(self.xref.file, self.hashstr, self.hashu32_0, self.hashu32_1)


class XrefPtr(ELFDissectStruct):
    fields = [
        ("xref", "P", Xref),
    ]


class XrefThreadSched(ELFDissectStruct, XrelfoJson):
    struct = "xref_threadsched"


Xref.containers[XREFT_EVENTSCHED] = XrefThreadSched


class XrefLogmsg(ELFDissectStruct, XrelfoJson):
    struct = "xref_logmsg"

    def _warn_fmt(self, text):
        lines = text.split("\n")
        yield (
            (self.xref.file, self.xref.line),
            "%s:%d: %s (in %s())%s\n"
            % (
                self.xref.file,
                self.xref.line,
                lines[0],
                self.xref.func,
                "".join(["\n" + l for l in lines[1:]]),
            ),
        )

    fmt_regexes = [
        (re.compile(r"([\n\t]+)"), "error: log message contains tab or newline"),
        #    (re.compile(r'^(\s+)'),   'warning: log message starts with whitespace'),
        (
            re.compile(r"^((?:warn(?:ing)?|error):\s*)", re.I),
            "warning: log message starts with severity",
        ),
    ]
    arg_regexes = [
        # the (?<![\?:] ) avoids warning for x ? inet_ntop(...) : "(bla)"
        (
            re.compile(r"((?<![\?:] )inet_ntop\s*\(\s*(?:[AP]F_INET|2)\s*,)"),
            "cleanup: replace inet_ntop(AF_INET, ...) with %pI4",
            lambda s: True,
        ),
        (
            re.compile(r"((?<![\?:] )inet_ntop\s*\(\s*(?:[AP]F_INET6|10)\s*,)"),
            "cleanup: replace inet_ntop(AF_INET6, ...) with %pI6",
            lambda s: True,
        ),
        (
            # string split-up here is to not trigger "inet_ntoa forbidden"
            re.compile(r"((?<![\?:] )inet_" + r"ntoa)"),
            "cleanup: replace inet_" + "ntoa(...) with %pI4",
            lambda s: True,
        ),
        (
            re.compile(r"((?<![\?:] )ipaddr2str)"),
            "cleanup: replace ipaddr2str(...) with %pIA",
            lambda s: True,
        ),
        (
            re.compile(r"((?<![\?:] )prefix2str)"),
            "cleanup: replace prefix2str(...) with %pFX",
            lambda s: True,
        ),
        (
            re.compile(r"((?<![\?:] )prefix_mac2str)"),
            "cleanup: replace prefix_mac2str(...) with %pEA",
            lambda s: True,
        ),
        (
            re.compile(r"((?<![\?:] )sockunion2str)"),
            "cleanup: replace sockunion2str(...) with %pSU",
            lambda s: True,
        ),
        #   (re.compile(r'^(\s*__(?:func|FUNCTION|PRETTY_FUNCTION)__\s*)'), 'error: debug message starts with __func__', lambda s: (s.priority & 7 == 7) ),
    ]

    def check(self, wopt):
        def fmt_msg(rex, itext):
            if sys.stderr.isatty():
                items = rex.split(itext)
                out = []
                for i, text in enumerate(items):
                    if (i % 2) == 1:
                        out.append("\033[41;37;1m%s\033[m" % repr(text)[1:-1])
                    else:
                        out.append(repr(text)[1:-1])

                excerpt = "".join(out)
            else:
                excerpt = repr(itext)[1:-1]
            return excerpt

        if wopt.Wlog_format:
            for rex, msg in self.fmt_regexes:
                if not rex.search(self.fmtstring):
                    continue

                excerpt = fmt_msg(rex, self.fmtstring)
                yield from self._warn_fmt('%s: "%s"' % (msg, excerpt))

        if wopt.Wlog_args:
            for rex, msg, cond in self.arg_regexes:
                if not cond(self):
                    continue
                if not rex.search(self.args):
                    continue

                excerpt = fmt_msg(rex, self.args)
                yield from self._warn_fmt(
                    '%s:\n\t"%s",\n\t%s' % (msg, repr(self.fmtstring)[1:-1], excerpt)
                )

    def dump(self):
        print(
            "%-60s %s%s %-25s [EC %d] %s"
            % (
                "%s:%d %s()" % (self.xref.file, self.xref.line, self.xref.func),
                prios[self.priority & 7],
                priovals.get(self.priority & 0x30, " "),
                self.xref.xrefdata.uid,
                self.ec,
                self.fmtstring,
            )
        )

    def to_dict(self, xrelfo):
        jsobj = dict([(i, getattr(self.xref, i)) for i in ["file", "line", "func"]])
        if self.ec != 0:
            jsobj["ec"] = self.ec
        jsobj["fmtstring"] = self.fmtstring
        jsobj["args"] = self.args
        jsobj["priority"] = self.priority & 7
        jsobj["type"] = "logmsg"
        jsobj["binary"] = self._elfsect._elfwrap.orig_filename

        if self.priority & 0x10:
            jsobj.setdefault("flags", []).append("errno")
        if self.priority & 0x20:
            jsobj.setdefault("flags", []).append("getaddrinfo")

        xrelfo["refs"].setdefault(self.xref.xrefdata.uid, []).append(jsobj)


Xref.containers[XREFT_LOGMSG] = XrefLogmsg


class CmdElement(ELFDissectStruct, XrelfoJson):
    struct = "cmd_element"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def to_dict(self, xrelfo):
        jsobj = (
            xrelfo["cli"]
            .setdefault(self.name, {})
            .setdefault(self._elfsect._elfwrap.orig_filename, {})
        )

        jsobj.update(
            {
                "string": self.string,
                "doc": self.doc,
            }
        )
        if self.attr:
            jsobj["attr"] = attr = self.attr
            for attrname in CmdAttr.__members__:
                val = CmdAttr[attrname]
                if attr & val:
                    jsobj.setdefault("attrs", []).append(attrname.lower())
                    attr &= ~val

        jsobj["defun"] = dict(
            [(i, getattr(self.xref, i)) for i in ["file", "line", "func"]]
        )


Xref.containers[XREFT_DEFUN] = CmdElement


class XrefInstallElement(ELFDissectStruct, XrelfoJson):
    struct = "xref_install_element"

    def to_dict(self, xrelfo):
        jsobj = (
            xrelfo["cli"]
            .setdefault(self.cmd_element.name, {})
            .setdefault(self._elfsect._elfwrap.orig_filename, {})
        )
        nodes = jsobj.setdefault("nodes", [])

        nodes.append(
            {
                "node": self.node_type,
                "install": dict(
                    [(i, getattr(self.xref, i)) for i in ["file", "line", "func"]]
                ),
            }
        )


Xref.containers[XREFT_INSTALL_ELEMENT] = XrefInstallElement

# shove in field defs
fieldapply = FieldApplicator(xrefstructs)
fieldapply.add(Xref)
fieldapply.add(Xrefdata)
fieldapply.add(XrefLogmsg)
fieldapply.add(XrefThreadSched)
fieldapply.add(CmdElement)
fieldapply.add(XrefInstallElement)
fieldapply()


class Xrelfo(dict):
    def __init__(self):
        super().__init__(
            {
                "refs": {},
                "cli": {},
            }
        )
        self._xrefs = []
        self.note_warn = False

    def load_file(self, filename):
        orig_filename = filename
        if filename.endswith(".la") or filename.endswith(".lo"):
            with open(filename, "r") as fd:
                for line in fd:
                    line = line.strip()
                    if line.startswith("#") or line == "" or "=" not in line:
                        continue

                    var, val = line.split("=", 1)
                    if var not in ["library_names", "pic_object"]:
                        continue
                    if val.startswith("'") or val.startswith('"'):
                        val = val[1:-1]

                    if var == "pic_object":
                        filename = os.path.join(os.path.dirname(filename), val)
                        break

                    val = val.strip().split()[0]
                    filename = os.path.join(os.path.dirname(filename), ".libs", val)
                    break
                else:
                    raise ValueError(
                        'could not process libtool file "%s"' % orig_filename
                    )

        while True:
            with open(filename, "rb") as fd:
                hdr = fd.read(4)

            if hdr == b"\x7fELF":
                self.load_elf(filename, orig_filename)
                return

            if hdr[:2] == b"#!":
                path, name = os.path.split(filename)
                filename = os.path.join(path, ".libs", name)
                continue

            if hdr[:1] == b"{":
                with open(filename, "r") as fd:
                    self.load_json(fd)
                return

            raise ValueError("cannot determine file type for %s" % (filename))

    def load_elf(self, filename, orig_filename):
        edf = ELFDissectFile(filename)
        edf.orig_filename = orig_filename

        note = edf._elffile.find_note("FRRouting", "XREF")
        if note is not None:
            endian = ">" if edf._elffile.bigendian else "<"
            mem = edf._elffile[note]
            if edf._elffile.elfclass == 64:
                start, end = struct.unpack(endian + "QQ", mem)
                start += note.start
                end += note.start + 8
            else:
                start, end = struct.unpack(endian + "II", mem)
                start += note.start
                end += note.start + 4

            ptrs = edf.iter_data(XrefPtr, slice(start, end))

        else:
            if elf_notes:
                self.note_warn = True
                sys.stderr.write(
                    """%s: warning: binary has no FRRouting.XREF note
%s-   one of FRR_MODULE_SETUP, FRR_DAEMON_INFO or XREF_SETUP must be used
"""
                    % (orig_filename, orig_filename)
                )

            xrefarray = edf.get_section("xref_array")
            if xrefarray is None:
                raise ValueError("file has neither xref note nor xref_array section")

            ptrs = xrefarray.iter_data(XrefPtr)

        for ptr in ptrs:
            if ptr.xref is None:
                print("NULL xref")
                continue
            self._xrefs.append(ptr.xref)

            container = ptr.xref.container()
            if container is None:
                continue
            container.to_dict(self)

        return edf

    def load_json(self, fd):
        data = json.load(fd)
        for uid, items in data["refs"].items():
            myitems = self["refs"].setdefault(uid, [])
            for item in items:
                if item in myitems:
                    continue
                myitems.append(item)

        for cmd, items in data["cli"].items():
            self["cli"].setdefault(cmd, {}).update(items)

        return data

    def check(self, checks):
        for xref in self._xrefs:
            yield from xref.check(checks)


def main():
    argp = argparse.ArgumentParser(description="FRR xref ELF extractor")
    argp.add_argument("-o", dest="output", type=str, help="write JSON output")
    argp.add_argument("--out-by-file", type=str, help="write by-file JSON output")
    argp.add_argument(
        "-c", dest="vtysh_cmds", type=str, help="write vtysh_cmd.c", nargs="*"
    )
    argp.add_argument("-Wlog-format", action="store_const", const=True)
    argp.add_argument("-Wlog-args", action="store_const", const=True)
    argp.add_argument("-Werror", action="store_const", const=True)
    argp.add_argument("--profile", action="store_const", const=True)
    argp.add_argument(
        "binaries",
        metavar="BINARY",
        nargs="+",
        type=str,
        help="files to read (ELF files or libtool objects)",
    )
    args = argp.parse_args()

    if args.profile:
        import cProfile

        cProfile.runctx("_main(args)", globals(), {"args": args}, sort="cumtime")
    else:
        _main(args)


def _main(args):
    errors = 0
    xrelfo = Xrelfo()

    for fn in args.binaries:
        try:
            xrelfo.load_file(fn)
        except:
            errors += 1
            sys.stderr.write("while processing %s:\n" % (fn))
            traceback.print_exc()

    if xrelfo.note_warn and args.Werror:
        errors += 1

    for option in dir(args):
        if option.startswith("W") and option != "Werror":
            checks = sorted(xrelfo.check(args))
            sys.stderr.write("".join([c[-1] for c in checks]))

            if args.Werror and len(checks) > 0:
                errors += 1
            break

    refs = xrelfo["refs"]

    counts = {}
    for k, v in refs.items():
        strs = set([i["fmtstring"] for i in v])
        if len(strs) != 1:
            print("\033[31;1m%s\033[m" % k)
        counts[k] = len(v)

    out = xrelfo
    outbyfile = {}
    for uid, locs in refs.items():
        for loc in locs:
            filearray = outbyfile.setdefault(loc["file"], [])
            loc = dict(loc)
            del loc["file"]
            filearray.append(loc)

    for k in outbyfile.keys():
        outbyfile[k] = sorted(outbyfile[k], key=lambda x: x["line"])

    if errors:
        sys.exit(1)

    if args.output:
        with open(args.output + ".tmp", "w") as fd:
            json.dump(out, fd, indent=2, sort_keys=True, **json_dump_args)
        os.rename(args.output + ".tmp", args.output)

    if args.out_by_file:
        with open(args.out_by_file + ".tmp", "w") as fd:
            json.dump(outbyfile, fd, indent=2, sort_keys=True, **json_dump_args)
        os.rename(args.out_by_file + ".tmp", args.out_by_file)

    if args.vtysh_cmds:
        fds = []
        for filename in args.vtysh_cmds:
            fds.append(open(filename + ".tmp", "w"))

        CommandEntry.run(out, fds)

        while fds:
            fds.pop(0).close()
        for filename in args.vtysh_cmds:
            os.rename(filename + ".tmp", filename)

        if args.Werror and CommandEntry.warn_counter:
            sys.exit(1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 eval: (blacken-mode 1) -*-
# SPDX-License-Identifier: GPL-2.0-or-later
#
# January 28 2023, Christian Hopps <chopps@labn.net>
#
# Copyright (c) 2023, LabN Consulting, L.L.C.
#
"""A tiny init for namespaces in python inspired by the C program tini."""


# pylint: disable=global-statement
import argparse
import errno
import logging
import os
import re
import select
import shlex
import signal
import subprocess
import sys

from signal import Signals as S


try:
    from munet import linux
except ModuleNotFoundError:
    # We cannot use relative imports and still run this module directly as a script, and
    # there are some use cases where we want to run this file as a script.
    sys.path.append(os.path.dirname(os.path.realpath(__file__)))
    import linux


class g:
    """Global variables for our program."""

    child_pid = -1
    orig_pid = os.getpid()
    exit_signal = False
    pid_status_cache = {}
    restore_signals = set()
    very_verbose = False


unshare_flags = {
    "C": linux.CLONE_NEWCGROUP,
    "i": linux.CLONE_NEWIPC,
    "m": linux.CLONE_NEWNS,
    "n": linux.CLONE_NEWNET,
    "p": linux.CLONE_NEWPID,
    "u": linux.CLONE_NEWUTS,
    "T": linux.CLONE_NEWTIME,
}


ignored_signals = {
    S.SIGTTIN,
    S.SIGTTOU,
}
abort_signals = {
    S.SIGABRT,
    S.SIGBUS,
    S.SIGFPE,
    S.SIGILL,
    S.SIGKILL,
    S.SIGSEGV,
    S.SIGSTOP,
    S.SIGSYS,
    S.SIGTRAP,
}
no_prop_signals = abort_signals | ignored_signals | {S.SIGCHLD}


def vdebug(*args, **kwargs):
    if g.very_verbose:
        logging.debug(*args, **kwargs)


def get_pid_status_item(status, stat):
    m = re.search(rf"(?:^|\n){stat}:\t(.*)(?:\n|$)", status)
    return m.group(1).strip() if m else None


def pget_pid_status_item(pid, stat):
    if pid not in g.pid_status_cache:
        with open(f"/proc/{pid}/status", "r", encoding="utf-8") as f:
            g.pid_status_cache[pid] = f.read().strip()
    return get_pid_status_item(g.pid_status_cache[pid], stat).strip()


def get_pid_name(pid):
    try:
        return get_pid_status_item(g.pid_status_cache[pid], "Name")
    except Exception:
        return str(pid)


# def init_get_child_pids():
#     """Return list of "children" pids.
#     We consider any process with a 0 parent pid to also be our child as it
#     nsentered our pid namespace from an external parent.
#     """
#     g.pid_status_cache.clear()
#     pids = (int(x) for x in os.listdir("/proc") if x.isdigit() and x != "1")
#     return (
#         x for x in pids if x == g.child_pid or pget_pid_status_item(x, "PPid") == "0"
#     )


def exit_with_status(status):
    if os.WIFEXITED(status):
        ec = os.WEXITSTATUS(status)
    elif os.WIFSIGNALED(status):
        ec = 0x80 | os.WTERMSIG(status)
    else:
        ec = 255
    logging.debug("exiting with code %s", ec)
    sys.exit(ec)


def __waitpid(tag, nohang=False):  # pylint: disable=inconsistent-return-statements
    if nohang:
        idobj = os.waitid(os.P_ALL, 0, os.WEXITED | os.WNOHANG)
        if idobj is None:
            return True
    else:
        idobj = os.waitid(os.P_ALL, 0, os.WEXITED)
        assert idobj is not None

    pid = idobj.si_pid
    status = idobj.si_status

    if pid != g.child_pid:
        pidname = get_pid_name(pid)
        logging.debug(
            "%s: reaped zombie %s (%s) w/ status %s", tag, pid, pidname, status
        )
        return False

    logging.debug("reaped child with status %s", status)
    exit_with_status(status)
    # NOTREACHED


def waitpid(tag):
    logging.debug("%s: waitid for exiting process", tag)
    __waitpid(tag, False)

    while True:
        logging.debug("%s: checking for another exiting process", tag)
        if __waitpid(tag, True):
            return


def sig_trasmit(signum, _):
    signame = signal.Signals(signum).name
    if g.child_pid == -1:
        # We've received a signal after setting up to be init proc
        # but prior to fork or fork returning with child pid
        logging.debug("received %s prior to child exec, exiting", signame)
        sys.exit(0x80 | signum)

    try:
        os.kill(g.child_pid, signum)
    except OSError as error:
        if error.errno != errno.ESRCH:
            logging.error(
                "error forwarding signal %s to child, exiting: %s", signum, error
            )
            sys.exit(0x80 | signum)
        logging.debug("child pid %s exited prior to signaling", g.child_pid)


def sig_sigchld(signum, _):
    assert signum == S.SIGCHLD


def setup_init_signals():
    valid = set(signal.valid_signals())
    named = set(x.value for x in signal.Signals)
    for snum in sorted(named):
        if snum not in valid:
            continue
        if S.SIGRTMIN <= snum <= S.SIGRTMAX:
            continue

        sname = signal.Signals(snum).name
        if snum == S.SIGCHLD:
            vdebug("installing local handler for %s", sname)
            signal.signal(snum, sig_sigchld)
            g.restore_signals.add(snum)
        elif snum in ignored_signals:
            vdebug("installing ignore handler for %s", sname)
            signal.signal(snum, signal.SIG_IGN)
            g.restore_signals.add(snum)
        elif snum in abort_signals:
            vdebug("leaving default handler for %s", sname)
            # signal.signal(snum, signal.SIG_DFL)
        else:
            vdebug("installing trasmit signal handler for %s", sname)
            try:
                signal.signal(snum, sig_trasmit)
                g.restore_signals.add(snum)
            except OSError as error:
                logging.warning(
                    "failed installing signal handler for %s: %s", sname, error
                )


def new_process_group():
    """Create and lead a new process group.

    This function will create a new process group if we are not yet leading one, and
    additionally foreground said process group in our session. This foregrounding
    action is copied from tini, and I believe serves a purpose when serving as init
    for a container (e.g., podman).
    """
    pid = os.getpid()
    try:
        pgid = os.getpgrp()
        if pgid == pid:
            logging.debug("already process group leader %s", pgid)
        else:
            logging.debug("creating new process group %s", pid)
            os.setpgid(pid, 0)
    except Exception as error:
        logging.warning("unable to get new process group: %s", error)
        return

    # Block these in order to allow foregrounding, otherwise we'd get SIGTTOU blocked
    signal.signal(S.SIGTTIN, signal.SIG_IGN)
    signal.signal(S.SIGTTOU, signal.SIG_IGN)
    fd = sys.stdin.fileno()
    if not os.isatty(fd):
        logging.debug("stdin not a tty no foregrounding required")
    else:
        try:
            # This will error if our session no longer associated with controlling tty.
            pgid = os.tcgetpgrp(fd)
            if pgid == pid:
                logging.debug("process group already in foreground %s", pgid)
            else:
                logging.debug("making us the foreground pgid backgrounding %s", pgid)
                os.tcsetpgrp(fd, pid)
        except OSError as error:
            if error.errno == errno.ENOTTY:
                logging.debug("session is no longer associated with controlling tty")
            else:
                logging.warning("unable to foreground pgid %s: %s", pid, error)
    signal.signal(S.SIGTTIN, signal.SIG_DFL)
    signal.signal(S.SIGTTOU, signal.SIG_DFL)


def is_creating_pid_namespace():
    p1name = subprocess.check_output(
        "readlink /proc/self/pid", stderr=subprocess.STDOUT, shell=True
    )
    p2name = subprocess.check_output(
        "readlink /proc/self/pid_for_children", stderr=subprocess.STDOUT, shell=True
    )
    return p1name != p2name


def poll_for_pids(msg, tag):
    poller = select.poll()
    while True:
        logging.info("%s", msg)
        events = poller.poll(1000)
        logging.info("init: poll: checking for zombies and child exit: %s", events)
        try:
            waitpid(tag)
        except ChildProcessError as error:
            logging.warning("init: got SIGCHLD but no pid to wait on: %s", error)
    # NOTREACHED


def be_init(new_pg, exec_args):
    #
    # Arrange for us to be killed when our parent dies, this will subsequently also kill
    # all procs in any PID namespace we are init for.
    #
    logging.debug("set us to be SIGKILLed when parent exits")
    linux.set_parent_death_signal(signal.SIGKILL)

    # If we are createing a new PID namespace for children...
    if g.orig_pid != 1:
        logging.debug("started as pid %s", g.orig_pid)
        # assert is_creating_pid_namespace()

        # Fork to become pid 1
        logging.debug("forking to become pid 1")
        child_pid = os.fork()
        if child_pid:
            logging.debug("in parent waiting on child pid %s to exit", child_pid)
            status = os.wait()
            logging.debug("got child exit status %s", status)
            exit_with_status(status)
            # NOTREACHED

        # We must be pid 1 now.
        logging.debug("in child as pid %s", os.getpid())
        assert os.getpid() == 1

        # We need a new /proc now.
        logging.debug("mount new /proc")
        linux.mount("proc", "/proc", "proc")

        # If the parent exists kill us using SIGKILL
        logging.debug("set us to be SIGKILLed when parent exits")
        linux.set_parent_death_signal(signal.SIGKILL)

    if not exec_args:
        if not new_pg:
            logging.debug("no exec args, no new process group")
            # # if 0 == os.getpgid(0):
            # status = os.setpgid(0, 1)
            # logging.debug("os.setpgid(0, 1) == %s", status)
        else:
            logging.debug("no exec args, creating new process group")
            # No exec so we are the "child".
            new_process_group()

        # Reap children as init process
        vdebug("installing local handler for SIGCHLD")
        signal.signal(signal.SIGCHLD, sig_sigchld)
        poll_for_pids("init: waiting to reap zombies", "PAUSE-EXIT")
        # NOTREACHED

    # Set (parent) signal handlers before any fork to avoid race
    setup_init_signals()

    logging.debug("forking to execute child")
    g.child_pid = os.fork()
    if g.child_pid == 0:
        # In child, restore signals to default handling:
        for snum in g.restore_signals:
            signal.signal(snum, signal.SIG_DFL)

        # XXX is a new pg right?
        new_process_group()
        logging.debug("child: executing '%s'", shlex.join(exec_args))
        os.execvp(exec_args[0], exec_args)
        # NOTREACHED

    poll_for_pids(f"parent: waiting for child pid {g.child_pid} to exit", "PARENT")
    # NOTREACHED


def unshare(flags):
    """Unshare into new namespaces."""
    uflags = 0
    for flag in flags:
        if flag not in unshare_flags:
            raise ValueError(f"unknown unshare flag '{flag}'")
        uflags |= unshare_flags[flag]
    new_pid = bool(uflags & linux.CLONE_NEWPID)
    new_mnt = bool(uflags & linux.CLONE_NEWNS)

    logging.debug("unshareing with flags: %s", linux.clone_flag_string(uflags))
    linux.unshare(uflags)

    if new_pid and not new_mnt:
        try:
            # If we are not creating new mount namspace, remount /proc private
            # so that our mount of a new /proc doesn't affect parent namespace
            logging.debug("remount /proc recursive private")
            linux.mount("none", "/proc", None, linux.MS_REC | linux.MS_PRIVATE)
        except OSError as error:
            # EINVAL is OK b/c /proc not mounted may cause an error
            if error.errno != errno.EINVAL:
                raise
    if new_mnt:
        # Remount root as recursive private.
        logging.debug("remount / recursive private")
        linux.mount("none", "/", None, linux.MS_REC | linux.MS_PRIVATE)

    # if new_pid:
    #     logging.debug("mount new /proc")
    #     linux.mount("proc", "/proc", "proc")

    return new_pid


def main():
    #
    # Parse CLI args.
    #

    ap = argparse.ArgumentParser()
    ap.add_argument(
        "-P",
        "--no-proc-group",
        action="store_true",
        help="set to inherit the process group",
    )
    valid_flags = "".join(unshare_flags)
    ap.add_argument(
        "--unshare-flags",
        help=(
            f"string of unshare(1) flags. Supported values from '{valid_flags}'."
            " 'm' will remount `/` recursive private. 'p' will remount /proc"
            " and fork, and the child will be signaled to exit on exit of parent.."
        ),
    )
    ap.add_argument(
        "-v", dest="verbose", action="count", default=0, help="more -v's, more verbose"
    )
    ap.add_argument("rest", nargs=argparse.REMAINDER)
    args = ap.parse_args()

    #
    # Setup logging.
    #

    level = logging.DEBUG if args.verbose else logging.INFO
    if args.verbose > 1:
        g.very_verbose = True
    logging.basicConfig(
        level=level, format="%(asctime)s mutini: %(levelname)s: %(message)s"
    )

    #
    # Run program
    #

    status = 5
    try:
        new_pid = False
        if args.unshare_flags:
            new_pid = unshare(args.unshare_flags)

        if g.orig_pid != 1 and not new_pid:
            # Simply hold the namespaces
            poll_for_pids("holding namespace waiting to be signaled to exit", "PARENT")
            # NOTREACHED

        be_init(not args.no_proc_group, args.rest)
        # NOTREACHED
        logging.critical("Exited from be_init!")
    except KeyboardInterrupt:
        logging.info("exiting (main), received KeyboardInterrupt in main")
        status = 0x80 | signal.SIGINT
    except Exception as error:
        logging.info("exiting (main), do to exception %s", error, exc_info=True)

    sys.exit(status)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 eval: (blacken-mode 1) -*-
# SPDX-License-Identifier: ISC
#
# June 10 2023, Christian Hopps <chopps@labn.net>
#
# Copyright (c) 2023, LabN Consulting, L.L.C.
#
"""
Test mgmtd parsing of configs.

So:

MGMTD matches zebra:

one exit file:    ONE: vty -f file
one exit redir:   ONE: vty < file
early exit file:  ONE: vty -f file
early exit redir: ONE: vty < file
early end file:   ALL: vty -f file
early end redir:  ONE: vty < file

Raw tests:

FAILED mgmt_config/test_config.py::test_mgmtd_one_exit_file - AssertionError: vtysh < didn't work after exit
FAILED mgmt_config/test_config.py::test_mgmtd_one_exit_redir - AssertionError: vtysh < didn't work after exit
FAILED mgmt_config/test_config.py::test_mgmtd_early_exit_file - AssertionError: vtysh -f didn't work after 1 exit
FAILED mgmt_config/test_config.py::test_mgmtd_early_exit_redir - AssertionError: vtysh < didn't work after 1 exits
FAILED mgmt_config/test_config.py::test_mgmtd_early_end_redir - AssertionError: vtysh < didn't work after 1 end

FAILED mgmt_config/test_config.py::test_zebra_one_exit_file - AssertionError: zebra second conf missing
FAILED mgmt_config/test_config.py::test_zebra_one_exit_redir - AssertionError: zebra second conf missing
FAILED mgmt_config/test_config.py::test_zebra_early_exit_file - AssertionError: zebra second conf missing
FAILED mgmt_config/test_config.py::test_zebra_early_exit_redir - AssertionError: zebra second conf missing
FAILED mgmt_config/test_config.py::test_zebra_early_end_redir - AssertionError: zebra second conf missing

Before fixed:

one exit file:    NONE: vty -f file
early exit file:  NONE: vty -f file

FAILED mgmt_config/test_config.py::test_mgmtd_one_exit_file - AssertionError: vtysh -f didn't work before exit
FAILED mgmt_config/test_config.py::test_mgmtd_one_exit_redir - AssertionError: vtysh < didn't work after exit
FAILED mgmt_config/test_config.py::test_mgmtd_early_exit_file - AssertionError: vtysh -f didn't work before exit
FAILED mgmt_config/test_config.py::test_mgmtd_early_exit_redir - AssertionError: vtysh < didn't work after 1 exits
FAILED mgmt_config/test_config.py::test_mgmtd_early_end_redir - AssertionError: vtysh < didn't work after 1 end

FAILED mgmt_config/test_config.py::test_zebra_one_exit_file - AssertionError: zebra second conf missing
FAILED mgmt_config/test_config.py::test_zebra_one_exit_redir - AssertionError: zebra second conf missing
FAILED mgmt_config/test_config.py::test_zebra_early_exit_file - AssertionError: zebra second conf missing
FAILED mgmt_config/test_config.py::test_zebra_early_exit_redir - AssertionError: zebra second conf missing
FAILED mgmt_config/test_config.py::test_zebra_early_end_redir - AssertionError: zebra second conf missing

"""
import ipaddress
import logging
import os
import re
from pathlib import Path

import pytest
from lib.common_config import retry, step
from lib.topogen import Topogen, TopoRouter

pytestmark = [pytest.mark.staticd, pytest.mark.mgmtd]


@retry(retry_timeout=1, initial_wait=0.1)
def check_kernel(r1, prefix, expected=True):
    net = ipaddress.ip_network(prefix)
    if net.version == 6:
        kernel = r1.cmd_nostatus("ip -6 route show", warn=not expected)
    else:
        kernel = r1.cmd_nostatus("ip -4 route show", warn=not expected)

    logging.debug("checking kernel routing table:\n%0.1920s", kernel)
    route = f"{str(net)}(?: nhid [0-9]+)?.*proto (static|196)"
    m = re.search(route, kernel)
    if expected and not m:
        return f"Failed to find \n'{route}'\n in \n'{kernel:.1920}'"
    elif not expected and m:
        return f"Failed found \n'{route}'\n in \n'{kernel:.1920}'"
    return None


@pytest.fixture(scope="module")
def tgen(request):
    "Setup/Teardown the environment and provide tgen argument to tests"

    topodef = {"s1": ("r1",)}

    tgen = Topogen(topodef, request.module.__name__)
    tgen.start_topology()

    # configure mgmtd using current mgmtd config file
    tgen.gears["r1"].load_config(TopoRouter.RD_ZEBRA, "zebra.conf")
    tgen.gears["r1"].load_config(TopoRouter.RD_MGMTD)

    tgen.start_router()
    yield tgen
    tgen.stop_topology()


def save_log_snippet(logfile, content, savepath=None):
    os.sync()
    os.sync()
    os.sync()

    with open(logfile, encoding="utf-8") as f:
        buf = f.read()
    assert content == buf[: len(content)]
    newcontent = buf[len(content) :]

    if savepath:
        with open(savepath, "w", encoding="utf-8") as f:
            f.write(newcontent)

    return buf


def mapname(lname):
    return lname.replace(".conf", "") + "-log.txt"


logbuf = ""


@pytest.fixture(scope="module")
def r1(tgen):
    return tgen.gears["r1"].net


@pytest.fixture(scope="module")
def confdir():
    return Path(os.environ["PYTEST_TOPOTEST_SCRIPTDIR"]) / "r1"


@pytest.fixture(scope="module")
def tempdir(r1):
    return Path(r1.rundir)


@pytest.fixture(scope="module")
def logpath(tempdir):
    return tempdir / "mgmtd.log"


@pytest.fixture(autouse=True, scope="function")
def cleanup_config(r1, tempdir, logpath):
    global logbuf

    logbuf = save_log_snippet(logpath, logbuf, "/dev/null")

    yield

    r1.cmd_nostatus("vtysh -c 'conf t' -c 'no allow-external-route-update'")
    r1.cmd_nostatus("vtysh -c 'conf t' -c 'no router-id 1.2.3.4'")
    r1.cmd_nostatus("vtysh -c 'conf t' -c 'no ip table range 2 3'")

    logbuf = save_log_snippet(logpath, logbuf, "/dev/null")


def test_staticd_startup(r1):
    r1.cmd_nostatus(
        "vtysh -c 'debug mgmt client frontend' "
        "-c 'debug mgmt client backend' "
        "-c 'debug mgmt backend frontend datastore transaction'"
    )
    step("Verifying routes are present on r1")
    result = check_kernel(r1, "12.0.0.0/24", retry_timeout=3.0)
    assert result is None


def test_mgmtd_one_exit_file(r1, confdir, tempdir, logpath):
    global logbuf

    conf = "one-exit.conf"
    step(f"load {conf} file with vtysh -f ")
    output = r1.cmd_nostatus(f"vtysh -f {confdir / conf}")
    logbuf = save_log_snippet(logpath, logbuf, tempdir / mapname(conf))
    print(output)

    result1 = check_kernel(r1, "20.1.0.0/24")
    result2 = check_kernel(r1, "20.2.0.0/24")

    assert result1 is None, "vtysh -f didn't work before exit"
    assert result2 is not None, "vtysh < worked after exit, unexpected"


def test_mgmtd_one_exit_redir(r1, confdir, tempdir, logpath):
    global logbuf

    conf = "one-exit2.conf"
    step(f"Redirect {conf} file into vtysh")
    output = r1.cmd_nostatus(f"vtysh < {confdir / conf}")
    logbuf = save_log_snippet(logpath, logbuf, tempdir / mapname(conf))
    print(output)

    result1 = check_kernel(r1, "21.1.0.0/24")
    result2 = check_kernel(r1, "21.2.0.0/24")

    assert result1 is None, "vtysh < didn't work before exit"
    assert result2 is not None, "vtysh < worked after exit, unexpected"


def test_mgmtd_early_exit_file(r1, confdir, tempdir, logpath):
    global logbuf

    conf = "early-exit.conf"
    step(f"load {conf} file with vtysh -f ")
    output = r1.cmd_nostatus(f"vtysh -f {confdir / conf}")
    logbuf = save_log_snippet(logpath, logbuf, tempdir / mapname(conf))
    print(output)

    result1 = check_kernel(r1, "13.1.0.0/24")
    result2 = check_kernel(r1, "13.2.0.0/24")
    result3 = check_kernel(r1, "13.3.0.0/24")

    assert result1 is None, "vtysh -f didn't work before exit"
    assert result2 is not None, "vtysh -f worked after 1 exit, unexpected"
    assert result3 is not None, "vtysh -f worked after 2 exit, unexpected"


def test_mgmtd_early_exit_redir(r1, confdir, tempdir, logpath):
    global logbuf

    conf = "early-exit2.conf"
    step(f"Redirect {conf} file into vtysh")
    output = r1.cmd_nostatus(f"vtysh < {confdir / conf}")
    logbuf = save_log_snippet(logpath, logbuf, tempdir / mapname(conf))
    print(output)

    result1 = check_kernel(r1, "14.1.0.0/24")
    result2 = check_kernel(r1, "14.2.0.0/24")
    result3 = check_kernel(r1, "14.3.0.0/24")

    assert result1 is None, "vtysh < didn't work before exit"
    assert result2 is not None, "vtysh < worked after 1 exits, unexpected"
    assert result3 is not None, "vtysh < worked after 2 exits, unexpected"


def test_mgmtd_early_end_file(r1, confdir, tempdir, logpath):
    global logbuf

    conf = "early-end.conf"
    step(f"load {conf} file with vtysh -f ")
    output = r1.cmd_nostatus(f"vtysh -f {confdir / conf}")
    logbuf = save_log_snippet(logpath, logbuf, tempdir / mapname(conf))
    print(output)

    result1 = check_kernel(r1, "15.1.0.0/24")
    result2 = check_kernel(r1, "15.2.0.0/24")
    result3 = check_kernel(r1, "15.3.0.0/24")

    assert result1 is None, "vtysh -f didn't work before end"
    assert result2 is None, "vtysh -f didn't work after 1 end"
    assert result3 is None, "vtysh -f didn't work after 2 ends"


def test_mgmtd_early_end_redir(r1, confdir, tempdir, logpath):
    global logbuf

    conf = "early-end2.conf"
    step(f"Redirect {conf} file into vtysh")
    output = r1.cmd_nostatus(f"vtysh < {confdir / conf}")
    logbuf = save_log_snippet(logpath, logbuf, tempdir / mapname(conf))
    print(output)

    result1 = check_kernel(r1, "16.1.0.0/24")
    result2 = check_kernel(r1, "16.2.0.0/24")
    result3 = check_kernel(r1, "16.3.0.0/24")

    assert result1 is None, "vtysh < didn't work before end"
    assert result2 is not None, "vtysh < worked after 1 end, unexpected"
    assert result3 is not None, "vtysh < worked after 2 end, unexpected"


#
# Zebra
#


def test_zebra_one_exit_file(r1, confdir, tempdir, logpath):
    global logbuf

    conf = "one-exit-zebra.conf"
    step(f"load {conf} file with vtysh -f ")
    output = r1.cmd_nostatus(f"vtysh -f {confdir / conf}")
    logbuf = save_log_snippet(logpath, logbuf, tempdir / mapname(conf))
    print(output)

    showrun = r1.cmd_nostatus("vtysh -c 'show running'")
    assert "allow-external-route-update" in showrun, "zebra conf missing"
    assert "router-id 1.2.3.4" not in showrun, "zebra second conf present, unexpected"


def test_zebra_one_exit_redir(r1, confdir, tempdir, logpath):
    global logbuf

    conf = "one-exit2-zebra.conf"
    step(f"Redirect {conf} file into vtysh")
    output = r1.cmd_nostatus(f"vtysh < {confdir / conf}")
    logbuf = save_log_snippet(logpath, logbuf, tempdir / mapname(conf))
    print(output)

    showrun = r1.cmd_nostatus("vtysh -c 'show running'")

    assert "allow-external-route-update" in showrun, "zebra conf missing"
    assert "router-id 1.2.3.4" not in showrun, "zebra second conf present, unexpected"


def test_zebra_early_exit_file(r1, confdir, tempdir, logpath):
    global logbuf

    conf = "early-exit-zebra.conf"
    step(f"load {conf} file with vtysh -f ")
    output = r1.cmd_nostatus(f"vtysh -f {confdir / conf}")
    logbuf = save_log_snippet(logpath, logbuf, tempdir / mapname(conf))
    print(output)

    showrun = r1.cmd_nostatus("vtysh -c 'show running'")

    assert "allow-external-route-update" in showrun, "zebra conf missing"
    assert "router-id 1.2.3.4" not in showrun, "zebra second conf present, unexpected"
    assert "ip table range 2 3" not in showrun, "zebra third conf present, unexpected"


def test_zebra_early_exit_redir(r1, confdir, tempdir, logpath):
    global logbuf

    conf = "early-exit2-zebra.conf"
    step(f"Redirect {conf} file into vtysh")
    output = r1.cmd_nostatus(f"vtysh < {confdir / conf}")
    logbuf = save_log_snippet(logpath, logbuf, tempdir / mapname(conf))
    print(output)

    showrun = r1.cmd_nostatus("vtysh -c 'show running'")

    assert "allow-external-route-update" in showrun, "zebra conf missing"
    assert "router-id 1.2.3.4" not in showrun, "zebra second conf present, unexpected"
    assert "ip table range 2 3" not in showrun, "zebra third conf present, unexpected"


def test_zebra_early_end_file(r1, confdir, tempdir, logpath):
    global logbuf

    conf = "early-end-zebra.conf"
    step(f"load {conf} file with vtysh -f ")
    output = r1.cmd_nostatus(f"vtysh -f {confdir / conf}")
    logbuf = save_log_snippet(logpath, logbuf, tempdir / mapname(conf))
    print(output)

    showrun = r1.cmd_nostatus("vtysh -c 'show running'")

    assert "allow-external-route-update" in showrun, "zebra conf missing"
    assert "router-id 1.2.3.4" in showrun, "zebra second conf missing"
    assert "ip table range 2 3" in showrun, "zebra third missing"


def test_zebra_early_end_redir(r1, confdir, tempdir, logpath):
    global logbuf

    conf = "early-end2-zebra.conf"
    step(f"Redirect {conf} file into vtysh")
    output = r1.cmd_nostatus(f"vtysh < {confdir / conf}")
    logbuf = save_log_snippet(logpath, logbuf, tempdir / mapname(conf))
    print(output)

    showrun = r1.cmd_nostatus("vtysh -c 'show running'")

    assert "allow-external-route-update" in showrun, "zebra conf missing"
    assert "router-id 1.2.3.4" not in showrun, "zebra second conf present, unexpected"
    assert "ip table range 2 3" not in showrun, "zebra third conf present, unexpected"



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# Copyright (c) 2022 by
# Donatas Abraitis <donatas@opensourcerouting.org>
#

"""
Check if we can override VPN underlay next-hop from PE1 to PE2.
"""

import os
import sys
import json
import pytest
import functools

CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
from lib import topotest
from lib.topogen import Topogen, TopoRouter, get_topogen

pytestmark = [pytest.mark.bgpd]


def build_topo(tgen):
    tgen.add_router("cpe1")
    tgen.add_router("cpe2")
    tgen.add_router("pe1")
    tgen.add_router("pe2")

    switch = tgen.add_switch("s1")
    switch.add_link(tgen.gears["cpe1"])
    switch.add_link(tgen.gears["pe1"])

    switch = tgen.add_switch("s2")
    switch.add_link(tgen.gears["pe1"])
    switch.add_link(tgen.gears["pe2"])

    switch = tgen.add_switch("s3")
    switch.add_link(tgen.gears["pe2"])
    switch.add_link(tgen.gears["cpe2"])


def setup_module(mod):
    tgen = Topogen(build_topo, mod.__name__)
    tgen.start_topology()

    pe1 = tgen.gears["pe1"]
    pe2 = tgen.gears["pe2"]

    pe1.run("ip link add RED type vrf table 1001")
    pe1.run("ip link set up dev RED")
    pe2.run("ip link add RED type vrf table 1001")
    pe2.run("ip link set up dev RED")
    pe1.run("ip link set pe1-eth0 master RED")
    pe2.run("ip link set pe2-eth1 master RED")

    pe1.run("sysctl -w net.ipv4.ip_forward=1")
    pe2.run("sysctl -w net.ipv4.ip_forward=1")
    pe1.run("sysctl -w net.mpls.conf.pe1-eth0.input=1")
    pe2.run("sysctl -w net.mpls.conf.pe2-eth1.input=1")

    router_list = tgen.routers()

    for _, (rname, router) in enumerate(router_list.items(), 1):
        router.load_config(
            TopoRouter.RD_ZEBRA, os.path.join(CWD, "{}/zebra.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_BGP, os.path.join(CWD, "{}/bgpd.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_OSPF6, os.path.join(CWD, "{}/ospf6d.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_LDP, os.path.join(CWD, "{}/ldpd.conf".format(rname))
        )

    tgen.start_router()


def teardown_module(mod):
    tgen = get_topogen()
    tgen.stop_topology()


def test_bgp_vpn_5549():
    tgen = get_topogen()

    pe2 = tgen.gears["pe2"]

    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    def _bgp_vpn_nexthop_changed():
        output = json.loads(pe2.vtysh_cmd("show bgp ipv4 vpn json"))
        expected = {
            "routes": {
                "routeDistinguishers": {
                    "192.168.1.2:2": {
                        "172.16.255.1/32": [
                            {"valid": True, "nexthops": [{"ip": "2001:db8::1"}]}
                        ],
                        "192.168.1.0/24": [
                            {"valid": True, "nexthops": [{"ip": "2001:db8:1::1"}]}
                        ],
                    }
                }
            }
        }
        return topotest.json_cmp(output, expected)

    def _bgp_verify_v4_nexthop_validity():
        output = json.loads(tgen.gears["cpe1"].vtysh_cmd("show bgp nexthop json"))
        expected = {
            "ipv4": {
                "192.168.1.2": {
                    "valid": True,
                    "complete": True,
                    "igpMetric": 0,
                    "pathCount": 0,
                    "nexthops": [{"interfaceName": "cpe1-eth0"}],
                },
            }
        }
        return topotest.json_cmp(output, expected)

    def _bgp_verify_v6_global_nexthop_validity():
        output = json.loads(tgen.gears["pe2"].vtysh_cmd("show bgp nexthop json"))
        expected = {
            "ipv6": {
                "2001:db8::1": {
                    "valid": True,
                    "complete": True,
                    "igpMetric": 0,
                    "pathCount": 2,
                    "nexthops": [{"interfaceName": "pe2-eth0"}],
                },
                "2001:db8:1::1": {
                    "valid": True,
                    "complete": True,
                    "igpMetric": 10,
                    "pathCount": 2,
                    "peer": "2001:db8:1::1",
                    "nexthops": [{"interfaceName": "pe2-eth0"}],
                },
            }
        }
        return topotest.json_cmp(output, expected)

    test_func = functools.partial(_bgp_vpn_nexthop_changed)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Failed overriding IPv6 next-hop for VPN underlay"

    test_func = functools.partial(_bgp_verify_v4_nexthop_validity)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "IPv4 nexthop is invalid"

    test_func = functools.partial(_bgp_verify_v6_global_nexthop_validity)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "IPv6 nexthop is invalid"


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



# SPDX-License-Identifier: GPL-2.0-or-later
# zlog fuzz-tester input generator
#
# Copyright (C) 2021  David Lamparter for NetDEF, Inc.

from itertools import chain
import struct

lengths = set([128])
# lengths = [[i, i + 1, i + 3, i - 1, i - 3] for i in lengths]
# lengths = set([i for i in chain(*lengths) if i >= 0])

dsts = [0, 1, 2, 3]
fmts = [0, 1, 2, 3]


def combo():
    for l0 in lengths:
        for l1 in lengths:
            for l2 in lengths:
                for fmt in fmts:
                    for dst in dsts:
                        yield (l0, l1, l2, fmt, dst)


for i, tup in enumerate(combo()):
    with open("input/i%d" % i, "wb") as fd:
        fd.write(struct.pack("HHHBB", *tup))



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# test_pim_cand_rp_bsr.py
#
# Copyright (c) 2024 ATCorp
# Jafar Al-Gharaibeh
#

import os
import sys
import pytest
import json
from functools import partial

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib import topotest
from lib.topogen import Topogen, get_topogen
from lib.topolog import logger
from lib.pim import verify_pim_rp_info
from lib.common_config import step, write_test_header, retry

from time import sleep

"""
test_pim_cand_rp_bsr.py: Test candidate RP/BSR functionality
"""

TOPOLOGY = """
    Candidate RP/BSR functionality

    +---+---+                      +---+---+
    | C-BSR |     10.0.0.0/24      | C-BSR |
    +  R1   + <--------+---------> +  R2   |
    |elected| .1       |        .2 |       |
    +---+---+          |           +---+---+
     .1 |              |  10.0.2.0/24  | .2
        | 10.0.1.0/24  |               |
     .3 |              +-----|  .4     | .4
    +---+---+                |---->+---+---+
    | C-RP  |    10.0.3.0/24       | C-RP  |
    +  R3   + <--------+---------> +  R4   |
    | prio  | .3       |        .4 |       |
    +---+---+          |           +---+---+
     .3 |              |              | .4
        |10.0.4.0/24   |   10.0.5.0/24|
     .5 |              |         .6   | .6
    +---+---+          +---------->+---+---+
    |       |                      |       |
    +  R5   + <------------------> +  R6   |
    |       | .5                .6 |       |
    +---+---+     10.0.6.0/24      +---+---+
"""

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# Required to instantiate the topology builder class.
pytestmark = [
    pytest.mark.pimd,
    pytest.mark.pim6d,
    pytest.mark.ospfd,
    pytest.mark.ospf6d,
]


def build_topo(tgen):
    "Build function"

    # Create 6 routers
    for rn in range(1, 7):
        tgen.add_router("r{}".format(rn))

    # Create 7 switches and connect routers
    sw1 = tgen.add_switch("s1")
    sw1.add_link(tgen.gears["r1"])
    sw1.add_link(tgen.gears["r2"])

    sw = tgen.add_switch("s2")
    sw.add_link(tgen.gears["r1"])
    sw.add_link(tgen.gears["r3"])

    sw = tgen.add_switch("s3")
    sw.add_link(tgen.gears["r2"])
    sw.add_link(tgen.gears["r4"])

    sw3 = tgen.add_switch("s4")
    sw3.add_link(tgen.gears["r3"])
    sw3.add_link(tgen.gears["r4"])

    sw = tgen.add_switch("s5")
    sw.add_link(tgen.gears["r3"])
    sw.add_link(tgen.gears["r5"])

    sw = tgen.add_switch("s6")
    sw.add_link(tgen.gears["r4"])
    sw.add_link(tgen.gears["r6"])

    sw = tgen.add_switch("s7")
    sw.add_link(tgen.gears["r5"])
    sw.add_link(tgen.gears["r6"])

    # make the diagnoal connections
    sw1.add_link(tgen.gears["r4"])
    sw3.add_link(tgen.gears["r6"])


def setup_module(mod):
    logger.info("PIM Candidate RP/BSR:\n {}".format(TOPOLOGY))

    tgen = Topogen(build_topo, mod.__name__)
    tgen.start_topology()

    router_list = tgen.routers()
    for rname, router in router_list.items():
        logger.info("Loading router %s" % rname)
        router.load_frr_config(os.path.join(CWD, "{}/frr.conf".format(rname)))

    # Initialize all routers.
    tgen.start_router()
    for router in router_list.values():
        if router.has_version("<", "4.0"):
            tgen.set_error("unsupported version")


def teardown_module(mod):
    "Teardown the pytest environment"
    tgen = get_topogen()
    tgen.stop_topology()


def test_pim_bsr_election_r1(request):
    "Test PIM BSR Election"
    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    if tgen.routers_have_failure():
        pytest.skip("skipped because of router(s) failure")

    r2 = tgen.gears["r2"]
    # r1 should be the BSR winner because it has higher priority
    expected = {
        "bsr": "10.0.0.1",
        "priority": 200,
        "state": "ACCEPT_PREFERRED",
    }

    test_func = partial(topotest.router_json_cmp, r2, "show ip pim bsr json", expected)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)

    assertmsg = "r2: r1 was not elected, bsr election mismatch"
    assert result is None, assertmsg


def test_pim_bsr_cand_bsr_r1(request):
    "Test PIM BSR candidate BSR"
    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    if tgen.routers_have_failure():
        pytest.skip("skipped because of router(s) failure")

    r2 = tgen.gears["r2"]

    # r2 is a candidate bsr with low priority: elected = False
    expected = {"address": "10.0.0.2", "priority": 100, "elected": False}
    test_func = partial(
        topotest.router_json_cmp, r2, "show ip pim bsr candidate-bsr json", expected
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)

    assertmsg = "r2: candidate bsr mismatch "
    assert result is None, assertmsg


def test_pim_bsr_cand_rp(request):
    "Test PIM BSR candidate RP"
    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    if tgen.routers_have_failure():
        pytest.skip("skipped because of router(s) failure")

    r3 = tgen.gears["r3"]

    # r3 is a candidate rp
    expected = {"address": "10.0.3.3", "priority": 10}
    test_func = partial(
        topotest.router_json_cmp, r3, "show ip pim bsr candidate-rp json", expected
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)

    assertmsg = "r3: bsr candidate rp mismatch"
    assert result is None, assertmsg


def test_pim_bsr_rp_info(request):
    "Test RP info state"
    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    if tgen.routers_have_failure():
        pytest.skip("skipped because of router(s) failure")

    # At this point, all nodes, including r5 should have synced the RP state
    step("Verify rp-info on r5 from BSR")
    result = verify_pim_rp_info(
        tgen,
        None,
        "r5",
        "239.0.0.0/16",
        None,
        "10.0.3.3",
        "BSR",
        False,
        "ipv4",
        True,
        retry_timeout=90,
    )
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    result = verify_pim_rp_info(
        tgen,
        None,
        "r5",
        "239.0.0.0/8",
        None,
        "10.0.3.4",
        "BSR",
        False,
        "ipv4",
        True,
        retry_timeout=30,
    )
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    result = verify_pim_rp_info(
        tgen,
        None,
        "r5",
        "239.0.0.0/24",
        None,
        "10.0.3.4",
        "BSR",
        False,
        "ipv4",
        True,
        retry_timeout=30,
    )
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("Verify rp-info on the BSR node itself r1")
    result = verify_pim_rp_info(
        tgen,
        None,
        "r1",
        "239.0.0.0/16",
        None,
        "10.0.3.3",
        "BSR",
        False,
        "ipv4",
        True,
        retry_timeout=10,
    )
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    result = verify_pim_rp_info(
        tgen,
        None,
        "r1",
        "239.0.0.0/8",
        None,
        "10.0.3.4",
        "BSR",
        False,
        "ipv4",
        True,
        retry_timeout=10,
    )
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    result = verify_pim_rp_info(
        tgen,
        None,
        "r1",
        "239.0.0.0/24",
        None,
        "10.0.3.4",
        "BSR",
        False,
        "ipv4",
        True,
        retry_timeout=10,
    )
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)


def test_pim_bsr_election_fallback_r2(request):
    "Test PIM BSR Election Backup"
    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    if tgen.routers_have_failure():
        pytest.skip("skipped because of router(s) failure")

    step("Take r1 out from BSR candidates")
    r1 = tgen.gears["r1"]
    r1.vtysh_cmd(
        """
        configure
          router pim
            no bsr candidate-bsr priority 200 source address 10.0.0.1
        """
    )

    step("Verify r1 is no longer a BSR candidate")
    expected = {}

    test_func = partial(
        topotest.router_json_cmp, r1, "show ip pim bsr candidate-bsr json", expected
    )
    _, result = topotest.run_and_expect(test_func, None, count=10, wait=1)

    assertmsg = "r1: failed to remove bsr candidate configuration"
    assert result is None, assertmsg

    r2 = tgen.gears["r2"]
    # We should fall back to r2 as the BSR
    expected = {
        "bsr": "10.0.0.2",
        "priority": 100,
        "state": "BSR_ELECTED",
    }

    step("Verify that we fallback to r2 as the new BSR")

    test_func = partial(topotest.router_json_cmp, r2, "show ip pim bsr json", expected)
    _, result = topotest.run_and_expect(test_func, None, count=180, wait=1)

    assertmsg = "r2: failed to fallback to r2 as a BSR"
    assert result is None, assertmsg


def test_pim_bsr_rp_info_fallback(request):
    "Test RP info state on r5"
    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    if tgen.routers_have_failure():
        pytest.skip("skipped because of router(s) failure")

    step("Take r3 out from RP candidates for group 239.0.0.0/16")
    r3 = tgen.gears["r3"]
    r3.vtysh_cmd(
        """
        configure
          router pim
            no bsr candidate-rp group 239.0.0.0/16
        """
    )

    step("Verify falling back to r4 as the new RP for 239.0.0.0/16")

    result = verify_pim_rp_info(
        tgen,
        None,
        "r5",
        "239.0.0.0/16",
        None,
        "10.0.3.4",
        "BSR",
        False,
        "ipv4",
        True,
        retry_timeout=30,
    )
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)


def test_pimv6_bsr_election_r1(request):
    "Test PIMv6 BSR Election"
    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    if tgen.routers_have_failure():
        pytest.skip("skipped because of router(s) failure")

    r2 = tgen.gears["r2"]
    # r1 should be the BSR winner because it has higher priority
    expected = {
        "bsr": "fd00::1",
        "priority": 200,
        "state": "ACCEPT_PREFERRED",
    }

    test_func = partial(
        topotest.router_json_cmp, r2, "show ipv6 pim bsr json", expected
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)

    assertmsg = "r2: r1 was not elected, IPv6 bsr election mismatch"
    assert result is None, assertmsg


def test_pimv6_bsr_cand_rp(request):
    "Test PIMv6 BSR candidate RP"
    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    if tgen.routers_have_failure():
        pytest.skip("skipped because of router(s) failure")

    r3 = tgen.gears["r3"]

    # r3 is a candidate rp
    expected = {"address": "fd00:0:0:3::3", "priority": 10}
    test_func = partial(
        topotest.router_json_cmp, r3, "show ipv6 pim bsr candidate-rp json", expected
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)

    assertmsg = "r3: bsr candidate rp mismatch"
    assert result is None, assertmsg


def test_pimv6_bsr_rp_info(request):
    "Test IPv6 RP info state"
    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    if tgen.routers_have_failure():
        pytest.skip("skipped because of router(s) failure")

    # At this point, all nodes, including r5 should have synced the RP state
    step("Verify rp-info on r5 from BSR")
    result = verify_pim_rp_info(
        tgen,
        None,
        "r5",
        "ffbb::0/64",
        None,
        "fd00:0:0:3::3",
        "BSR",
        False,
        "ipv6",
        True,
        retry_timeout=90,
    )
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    result = verify_pim_rp_info(
        tgen,
        None,
        "r5",
        "ffbb::0/124",
        None,
        "fd00:0:0:3::4",
        "BSR",
        False,
        "ipv6",
        True,
        retry_timeout=30,
    )
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    result = verify_pim_rp_info(
        tgen,
        None,
        "r5",
        "ffbb::0/108",
        None,
        "fd00:0:0:3::4",
        "BSR",
        False,
        "ipv6",
        True,
        retry_timeout=30,
    )
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)


def test_memory_leak():
    "Run the memory leak test and report results."
    tgen = get_topogen()
    if not tgen.is_memleak_enabled():
        pytest.skip("Memory leak test/report is disabled")

    tgen.report_memory_leaks()


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# Copyright (c) 2019 by VMware, Inc. ("VMware")
# Used Copyright (c) 2018 by Network Device Education Foundation, Inc.
# ("NetDEF") in this file.
#


"""
Following tests are covered to test ecmp functionality on iBGP.
1. Verify bgp fast-convergence functionality
"""
import os
import sys
import time
import pytest
import re
from time import sleep

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))
sys.path.append(os.path.join(CWD, "../../"))

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib.topogen import Topogen, get_topogen

from lib.common_config import (
    write_test_header,
    write_test_footer,
    verify_rib,
    create_static_routes,
    check_address_types,
    reset_config_on_routers,
    shutdown_bringup_interface,
    apply_raw_config,
    start_topology,
)
from lib.topolog import logger
from lib.topojson import build_config_from_json
from lib.bgp import create_router_bgp, verify_bgp_convergence

pytestmark = [pytest.mark.bgpd, pytest.mark.staticd]


# Global variables
NEXT_HOPS = {"ipv4": [], "ipv6": []}
NETWORK = {"ipv4": "192.168.1.10/32", "ipv6": "fd00:0:0:1::10/128"}
NEXT_HOP_IP = {"ipv4": "10.0.0.1", "ipv6": "fd00::1"}
BGP_CONVERGENCE = False


def setup_module(mod):
    """
    Sets up the pytest environment.

    * `mod`: module name
    """
    global ADDR_TYPES

    testsuite_run_time = time.asctime(time.localtime(time.time()))
    logger.info("Testsuite start time: {}".format(testsuite_run_time))
    logger.info("=" * 40)

    # This function initiates the topology build with Topogen...
    json_file = "{}/ibgp_ecmp_topo3.json".format(CWD)
    tgen = Topogen(json_file, mod.__name__)
    global topo
    topo = tgen.json_topo

    # Starting topology, create tmp files which are loaded to routers
    #  to start daemons and then start routers
    start_topology(tgen)

    # Creating configuration from JSON
    build_config_from_json(tgen, topo)

    # Don't run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Api call verify whether BGP is converged
    ADDR_TYPES = check_address_types()

    BGP_CONVERGENCE = verify_bgp_convergence(tgen, topo)
    assert BGP_CONVERGENCE is True, "setup_module :Failed \n Error:" " {}".format(
        BGP_CONVERGENCE
    )

    # STATIC_ROUTE = True
    logger.info("Running setup_module() done")


def teardown_module():
    get_topogen().stop_topology()


def static_or_nw(tgen, topo, tc_name, test_type, dut):

    if test_type == "redist_static":
        input_dict_static = {
            dut: {
                "static_routes": [
                    {"network": NETWORK["ipv4"], "next_hop": NEXT_HOP_IP["ipv4"]},
                    {"network": NETWORK["ipv6"], "next_hop": NEXT_HOP_IP["ipv6"]},
                ]
            }
        }
        logger.info("Configuring static route on router %s", dut)
        result = create_static_routes(tgen, input_dict_static)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

        input_dict_2 = {
            dut: {
                "bgp": {
                    "address_family": {
                        "ipv4": {
                            "unicast": {"redistribute": [{"redist_type": "static"}]}
                        },
                        "ipv6": {
                            "unicast": {"redistribute": [{"redist_type": "static"}]}
                        },
                    }
                }
            }
        }

        logger.info("Configuring redistribute static route on router %s", dut)
        result = create_router_bgp(tgen, topo, input_dict_2)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    elif test_type == "advertise_nw":
        input_dict_nw = {
            dut: {
                "bgp": {
                    "address_family": {
                        "ipv4": {
                            "unicast": {
                                "advertise_networks": [{"network": NETWORK["ipv4"]}]
                            }
                        },
                        "ipv6": {
                            "unicast": {
                                "advertise_networks": [{"network": NETWORK["ipv6"]}]
                            }
                        },
                    }
                }
            }
        }

        logger.info(
            "Advertising networks %s %s from router %s",
            NETWORK["ipv4"],
            NETWORK["ipv6"],
            dut,
        )
        result = create_router_bgp(tgen, topo, input_dict_nw)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )


@pytest.mark.parametrize("test_type", ["redist_static"])
def test_ecmp_fast_convergence(request, test_type, tgen, topo):
    """This test is to verify bgp fast-convergence cli functionality"""

    tc_name = request.node.name
    write_test_header(tc_name)

    # Verifying RIB routes
    dut = "r3"
    protocol = "bgp"

    reset_config_on_routers(tgen)
    static_or_nw(tgen, topo, tc_name, test_type, "r2")

    for addr_type in ADDR_TYPES:
        input_dict = {"r3": {"static_routes": [{"network": NETWORK[addr_type]}]}}

        logger.info("Verifying %s routes on r3", addr_type)
        result = verify_rib(
            tgen,
            addr_type,
            dut,
            input_dict,
            protocol=protocol,
        )
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    intf1 = topo["routers"]["r2"]["links"]["r3-link1"]["interface"]
    intf2 = topo["routers"]["r2"]["links"]["r3-link2"]["interface"]

    logger.info("Shutdown one of the link b/w r2 and r3")
    shutdown_bringup_interface(tgen, "r2", intf1, False)

    logger.info("Verify bgp neighbors are still up")
    result = verify_bgp_convergence(tgen, topo)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    logger.info("Shutdown another link b/w r2 and r3")
    shutdown_bringup_interface(tgen, "r2", intf2, False)

    logger.info("Wait for 10 sec and make sure bgp neighbors are still up")
    sleep(10)
    result = verify_bgp_convergence(tgen, topo)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    logger.info("No shut links b/w r2 and r3")
    shutdown_bringup_interface(tgen, "r2", intf1, True)
    shutdown_bringup_interface(tgen, "r2", intf2, True)

    logger.info("Ensure that the links are still up")
    result = verify_bgp_convergence(tgen, topo)

    logger.info("Enable bgp fast-convergence cli")
    raw_config = {
        "r2": {
            "raw_config": [
                "router bgp {}".format(topo["routers"]["r2"]["bgp"]["local_as"]),
                "bgp fast-convergence",
            ]
        }
    }
    result = apply_raw_config(tgen, raw_config)
    assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    logger.info("Ensure BGP has processed the cli")
    r2 = tgen.gears["r2"]
    output = r2.vtysh_cmd("show run")
    verify = re.search(r"fast-convergence", output)
    assert verify is not None, "r2 does not have the fast convergence command yet"

    logger.info("Shutdown one link b/w r2 and r3")
    shutdown_bringup_interface(tgen, "r2", intf1, False)

    logger.info("Verify bgp neighbors goes down immediately")
    result = verify_bgp_convergence(tgen, topo, dut="r2", expected=False)
    assert result is not True, (
        "Testcase {} : Failed \n "
        "Expected: BGP should not be converged for {} \n "
        "Found: {}".format(tc_name, "r2", result)
    )

    logger.info("Shutdown second link b/w r2 and r3")
    shutdown_bringup_interface(tgen, "r2", intf2, False)

    logger.info("Verify bgp neighbors goes down immediately")
    result = verify_bgp_convergence(tgen, topo, dut="r2", expected=False)
    assert result is not True, (
        "Testcase {} : Failed \n "
        "Expected: BGP should not be converged for {} \n "
        "Found: {}".format(tc_name, "r2", result)
    )

    write_test_footer(tc_name)


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/python
# SPDX-License-Identifier: ISC
#
# test_bgp_roles_filtering.py
# Part of NetDEF Topology Tests
#
# Copyright (c) 2022 by Eugene Bogomazov <eb@qrator.net>
# Copyright (c) 2017 by
# Network Device Education Foundation, Inc. ("NetDEF")
#

"""
test_bgp_roles_filtering: test leaks prevention and mitigation with roles
"""

import json
import os
import sys
import pytest

CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
from lib import topotest
from lib.topogen import Topogen, TopoRouter

pytestmark = [pytest.mark.bgpd]


topodef = {f"s{i}": (f"r{i}", "r10") for i in range(1, 8)}


@pytest.fixture(scope="module")
def tgen(request):
    tgen = Topogen(topodef, request.module.__name__)
    tgen.start_topology()
    router_list = tgen.routers()
    for _, router in router_list.items():
        router.load_config(TopoRouter.RD_ZEBRA, "zebra.conf")
        router.load_config(TopoRouter.RD_BGP, "bgpd.conf")
    tgen.start_router()
    yield tgen
    tgen.stop_topology()


@pytest.fixture(autouse=True)
def skip_on_failure(tgen):
    if tgen.routers_have_failure():
        pytest.skip("skipped because of previous test failure")


def test_r10_routes(tgen):
    # provider-undefine pair bur strict-mode was set
    def _routes_half_converged():
        routes = json.loads(tgen.gears["r10"].vtysh_cmd("show bgp ipv4 json"))["routes"]
        output = sorted(routes.keys())
        expected = [
            "192.0.2.1/32",
            "192.0.2.2/32",
            "192.0.2.3/32",
            "192.0.2.4/32",
            "192.0.2.5/32",
            "192.0.2.6/32",
            "192.0.2.7/32",
        ]
        return output == expected

    success, _ = topotest.run_and_expect(_routes_half_converged, True, count=20, wait=3)
    assert success, "Routes did not converged"

    routes_with_otc = list()
    for number in range(1, 8):
        prefix = f"192.0.2.{number}/32"
        route_details = json.loads(
            tgen.gears["r10"].vtysh_cmd(f"show bgp ipv4 {prefix} json")
        )
        if route_details["paths"][0].get("otc") is not None:
            routes_with_otc.append(prefix)
    assert routes_with_otc == [
        "192.0.2.1/32",
        "192.0.2.2/32",
        "192.0.2.6/32",
        "192.0.2.7/32",
    ]


def test_r1_routes(tgen):
    routes = json.loads(tgen.gears["r1"].vtysh_cmd("show bgp ipv4 json"))["routes"]
    routes_list = sorted(routes.keys())
    assert routes_list == [
        "192.0.2.1/32",  # own
        "192.0.2.3/32",
        "192.0.2.4/32",
        "192.0.2.5/32",
    ]


def test_r6_routes(tgen):
    routes = json.loads(tgen.gears["r6"].vtysh_cmd("show bgp ipv4 json"))["routes"]
    routes_list = sorted(routes.keys())
    assert routes_list == [
        "192.0.2.3/32",
        "192.0.2.4/32",
        "192.0.2.5/32",
        "192.0.2.6/32",  # own
    ]


def test_r4_routes(tgen):
    routes = json.loads(tgen.gears["r4"].vtysh_cmd("show bgp ipv4 json"))["routes"]
    routes_list = sorted(routes.keys())
    assert routes_list == [
        "192.0.2.1/32",
        "192.0.2.2/32",
        "192.0.2.3/32",
        "192.0.2.4/32",
        "192.0.2.5/32",
        "192.0.2.6/32",
        "192.0.2.7/32",
    ]


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# test_bgp_reject_as_sets.py
# Part of NetDEF Topology Tests
#
# Copyright (c) 2019 by
# Donatas Abraitis <donatas.abraitis@gmail.com>
#

"""
Test if an aggregated route with AS_SET is not sent to peers.
Addressing draft-ietf-idr-deprecate-as-set-confed-set recommendations.

BGP speakers conforming to this document (i.e., conformant BGP
   speakers) MUST NOT locally generate BGP UPDATE messages containing
   AS_SET or AS_CONFED_SET.  Conformant BGP speakers SHOULD NOT send BGP
   UPDATE messages containing AS_SET or AS_CONFED_SET.  Upon receipt of
   such messages, conformant BGP speakers SHOULD use the "Treat-as-
   withdraw" error handling behavior as per [RFC7606].
"""

import os
import sys
import json
import pytest
import functools

CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
from lib import topotest
from lib.topogen import Topogen, TopoRouter, get_topogen

pytestmark = [pytest.mark.bgpd]


def build_topo(tgen):
    for routern in range(1, 4):
        tgen.add_router("r{}".format(routern))

    switch = tgen.add_switch("s1")
    switch.add_link(tgen.gears["r1"])
    switch.add_link(tgen.gears["r2"])

    switch = tgen.add_switch("s2")
    switch.add_link(tgen.gears["r2"])
    switch.add_link(tgen.gears["r3"])


def setup_module(mod):
    tgen = Topogen(build_topo, mod.__name__)
    tgen.start_topology()

    router_list = tgen.routers()

    for _, (rname, router) in enumerate(router_list.items(), 1):
        router.load_config(
            TopoRouter.RD_ZEBRA, os.path.join(CWD, "{}/zebra.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_BGP, os.path.join(CWD, "{}/bgpd.conf".format(rname))
        )

    tgen.start_router()


def teardown_module(mod):
    tgen = get_topogen()
    tgen.stop_topology()


def test_bgp_reject_as_sets():
    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    router = tgen.gears["r2"]

    def _bgp_converge(router):
        output = json.loads(router.vtysh_cmd("show ip bgp neighbor 192.168.255.2 json"))
        expected = {
            "192.168.255.2": {
                "bgpState": "Established",
                "addressFamilyInfo": {"ipv4Unicast": {"acceptedPrefixCounter": 2}},
            }
        }
        return topotest.json_cmp(output, expected)

    def _bgp_has_aggregated_route_with_stripped_as_set(router):
        output = json.loads(router.vtysh_cmd("show ip bgp 172.16.0.0/16 json"))
        expected = {
            "paths": [{"aspath": {"string": "Local", "segments": [], "length": 0}}]
        }
        return topotest.json_cmp(output, expected)

    def _bgp_announce_route_without_as_sets(router):
        output = json.loads(
            router.vtysh_cmd(
                "show ip bgp neighbor 192.168.254.2 advertised-routes json"
            )
        )
        expected = {
            "advertisedRoutes": {
                "172.16.0.0/16": {"path": ""},
                "192.168.254.0/30": {"path": "65003"},
                "192.168.255.0/30": {"path": "65001"},
            },
            "totalPrefixCounter": 3,
        }
        return topotest.json_cmp(output, expected)

    test_func = functools.partial(_bgp_converge, router)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)

    assert result is None, 'Failed bgp convergence in "{}"'.format(router)

    test_func = functools.partial(
        _bgp_has_aggregated_route_with_stripped_as_set, router
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)

    assert result is None, 'Failed to see an aggregated route in "{}"'.format(router)

    test_func = functools.partial(_bgp_announce_route_without_as_sets, router)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)

    assert (
        result is None
    ), 'Route 172.16.0.0/16 should be sent without AS_SET to r3 "{}"'.format(router)


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



from lib.lutil import luCommand

luCommand("r1", 'vtysh -c "show bgp ipv4 vpn"', "", "none", "VPN SAFI")
luCommand("r2", 'vtysh -c "show bgp ipv4 vpn"', "", "none", "VPN SAFI")
luCommand("r3", 'vtysh -c "show bgp ipv4 vpn"', "", "none", "VPN SAFI")
luCommand("r4", 'vtysh -c "show bgp ipv4 vpn"', "", "none", "VPN SAFI")
luCommand(
    "r1",
    'vtysh -c "show vnc registrations"',
    "Locally: *Active:  1 .* Remotely: *Active:  3",
    "wait",
    "See all registrations",
)
luCommand(
    "r3",
    'vtysh -c "show vnc registrations"',
    "Locally: *Active:  1 .* Remotely: *Active:  3",
    "wait",
    "See all registrations",
)
luCommand(
    "r4",
    'vtysh -c "show vnc registrations"',
    "Locally: *Active:  2 .* Remotely: *Active:  2",
    "wait",
    "See all registrations",
)
num = "4 routes and 4"
luCommand("r1", 'vtysh -c "show bgp ipv4 vpn"', num, "pass", "VPN SAFI okay")
luCommand("r2", 'vtysh -c "show bgp ipv4 vpn"', num, "pass", "VPN SAFI okay")
luCommand("r3", 'vtysh -c "show bgp ipv4 vpn"', num, "pass", "VPN SAFI okay")
luCommand("r4", 'vtysh -c "show bgp ipv4 vpn"', num, "pass", "VPN SAFI okay")
luCommand(
    "r1",
    'vtysh -c "debug rfapi-dev query vn 10.0.0.1 un 1.1.1.1 target 22.22.22.22"',
    "pfx=",
    "pass",
    "Query R2s info",
)
luCommand(
    "r1",
    'vtysh -c "debug rfapi-dev query vn 10.0.0.1 un 1.1.1.1 target 33.33.33.33"',
    "pfx=",
    "pass",
    "Query R4s info",
)
luCommand(
    "r3",
    'vtysh -c "debug rfapi-dev query vn 10.0.0.2 un 2.2.2.2 target 11.11.11.11"',
    "11.11.11.0/24.*11.11.11.0/24.*",
    "pass",
    "Query R1s+R4s info",
)
luCommand(
    "r3",
    'vtysh -c "debug rfapi-dev query vn 10.0.0.2 un 2.2.2.2 target 33.33.33.33"',
    "pfx=",
    "pass",
    "Query R4s info",
)
luCommand(
    "r4",
    'vtysh -c "debug rfapi-dev query vn 10.0.0.3 un 3.3.3.3 target 11.11.11.11"',
    "11.11.11.0/24.*11.11.11.0/24.*",
    "pass",
    "Query R1s+R4s info",
)
luCommand(
    "r4",
    'vtysh -c "debug rfapi-dev query vn 10.0.0.3 un 3.3.3.3 target 22.22.22.22"',
    "pfx=",
    "pass",
    "Query R2s info",
)



#!/usr/bin/python3
# SPDX-License-Identifier: GPL-2.0-or-later

# Copyright (C) 2023 Tomas Hlavacek (tmshlvck@gmail.com)

from typing import List, Tuple, Callable, Type
import socket
import threading
import socketserver
import struct
import ipaddress
import csv
import os
import sys

LISTEN_HOST, LISTEN_PORT = "0.0.0.0", 15432
VRPS_FILE = os.path.join(sys.path[0], "vrps.csv")


def dbg(m: str):
    print(m)
    sys.stdout.flush()


class RTRDatabase(object):
    def __init__(self, vrps_file: str) -> None:
        self.last_serial = 0
        self.ann4 = []
        self.ann6 = []
        self.withdraw4 = []
        self.withdraw6 = []

        with open(vrps_file, "r") as fh:
            for rasn, rnet, rmaxlen, _ in csv.reader(fh):
                try:
                    net = ipaddress.ip_network(rnet)
                    asn = int(rasn[2:])
                    maxlen = int(rmaxlen)
                    if net.version == 6:
                        self.ann6.append((asn, str(net), maxlen))
                    elif net.version == 4:
                        self.ann4.append((asn, str(net), maxlen))
                    else:
                        raise ValueError(f"Unknown AFI: {net.version}")
                except Exception as e:
                    dbg(
                        f"VRPS load: ignoring {str((rasn, rnet,rmaxlen))} because {str(e)}"
                    )

    def get_serial(self) -> int:
        return self.last_serial

    def set_serial(self, serial: int) -> None:
        self.last_serial = serial

    def get_announcements4(self, serial: int = 0) -> List[Tuple[int, str, int]]:
        if serial > self.last_serial:
            return self.ann4
        else:
            return []

    def get_withdrawals4(self, serial: int = 0) -> List[Tuple[int, str, int]]:
        if serial > self.last_serial:
            return self.withdraw4
        else:
            return []

    def get_announcements6(self, serial: int = 0) -> List[Tuple[int, str, int]]:
        if serial > self.last_serial:
            return self.ann6
        else:
            return []

    def get_withdrawals6(self, serial: int = 0) -> List[Tuple[int, str, int]]:
        if serial > self.last_serial:
            return self.withdraw6
        else:
            return []


class RTRConnHandler(socketserver.BaseRequestHandler):
    PROTO_VERSION = 0

    def setup(self) -> None:
        self.session_id = 2345
        self.serial = 1024

        dbg(f"New connection from: {str(self.client_address)} ")
        # TODO: register for notifies

    def finish(self) -> None:
        pass
        # TODO: de-register

    HEADER_LEN = 8

    def decode_header(self, buf: bytes) -> Tuple[int, int, int, int]:
        # common header in all received packets
        return struct.unpack("!BBHI", buf)
        # reutnrs (proto_ver, pdu_type, sess_id, length)

    SERNOTIFY_TYPE = 0
    SERNOTIFY_LEN = 12

    def send_sernotify(self, serial: int) -> None:
        # serial notify PDU
        dbg(f"<Serial Notify session_id={self.session_id} serial={serial}")
        self.request.send(
            struct.pack(
                "!BBHII",
                self.PROTO_VERSION,
                self.SERNOTIFY_TYPE,
                self.session_id,
                self.SERNOTIFY_LEN,
                serial,
            )
        )

    CACHERESPONSE_TYPE = 3
    CACHERESPONSE_LEN = 8

    def send_cacheresponse(self) -> None:
        # cache response PDU
        dbg(f"<Cache response session_id={self.session_id}")
        self.request.send(
            struct.pack(
                "!BBHI",
                self.PROTO_VERSION,
                self.CACHERESPONSE_TYPE,
                self.session_id,
                self.CACHERESPONSE_LEN,
            )
        )

    FLAGS_ANNOUNCE = 1
    FLAGS_WITHDRAW = 0

    IPV4_TYPE = 4
    IPV4_LEN = 20

    def send_ipv4(self, ipnet: str, asn: int, maxlen: int, flags: int):
        # IPv4 PDU
        dbg(f"<IPv4 net={ipnet} asn={asn} maxlen={maxlen} flags={flags}")
        ip = ipaddress.IPv4Network(ipnet)
        self.request.send(
            struct.pack(
                "!BBHIBBBB4sI",
                self.PROTO_VERSION,
                self.IPV4_TYPE,
                0,
                self.IPV4_LEN,
                flags,
                ip.prefixlen,
                maxlen,
                0,
                ip.network_address.packed,
                asn,
            )
        )

    def announce_ipv4(self, ipnet, asn, maxlen):
        self.send_ipv4(ipnet, asn, maxlen, self.FLAGS_ANNOUNCE)

    def withdraw_ipv4(self, ipnet, asn, maxlen):
        self.send_ipv4(ipnet, asn, maxlen, self.FLAGS_WITHDRAW)

    IPV6_TYPE = 6
    IPV6_LEN = 32

    def send_ipv6(self, ipnet: str, asn: int, maxlen: int, flags: int):
        # IPv6 PDU
        dbg(f"<IPv6 net={ipnet} asn={asn} maxlen={maxlen} flags={flags}")
        ip = ipaddress.IPv6Network(ipnet)
        self.request.send(
            struct.pack(
                "!BBHIBBBB16sI",
                self.PROTO_VERSION,
                self.IPV6_TYPE,
                0,
                self.IPV6_LEN,
                flags,
                ip.prefixlen,
                maxlen,
                0,
                ip.network_address.packed,
                asn,
            )
        )

    def announce_ipv6(self, ipnet: str, asn: int, maxlen: int):
        self.send_ipv6(ipnet, asn, maxlen, self.FLAGS_ANNOUNCE)

    def withdraw_ipv6(self, ipnet: str, asn: int, maxlen: int):
        self.send_ipv6(ipnet, asn, maxlen, self.FLAGS_WITHDRAW)

    EOD_TYPE = 7
    EOD_LEN = 12

    def send_endofdata(self, serial: int):
        # end of data PDU
        dbg(f"<End of Data session_id={self.session_id} serial={serial}")
        self.server.db.set_serial(serial)
        self.request.send(
            struct.pack(
                "!BBHII",
                self.PROTO_VERSION,
                self.EOD_TYPE,
                self.session_id,
                self.EOD_LEN,
                serial,
            )
        )

    CACHERESET_TYPE = 8
    CACHERESET_LEN = 8

    def send_cachereset(self):
        # cache reset PDU
        dbg("<Cache Reset")
        self.request.send(
            struct.pack(
                "!BBHI",
                self.PROTO_VERSION,
                self.CACHERESET_TYPE,
                0,
                self.CACHERESET_LEN,
            )
        )

    SERIAL_QUERY_TYPE = 1
    SERIAL_QUERY_LEN = 12

    def handle_serial_query(self, buf: bytes, sess_id: int):
        serial = struct.unpack("!I", buf)[0]
        dbg(f">Serial query: {serial}")
        if sess_id:
            self.server.db.set_serial(serial)
        else:
            self.server.db.set_serial(0)
        self.send_cacheresponse()

        for asn, ipnet, maxlen in self.server.db.get_announcements4(serial):
            self.announce_ipv4(ipnet, asn, maxlen)

        for asn, ipnet, maxlen in self.server.db.get_withdrawals4(serial):
            self.withdraw_ipv4(ipnet, asn, maxlen)

        for asn, ipnet, maxlen in self.server.db.get_announcements6(serial):
            self.announce_ipv6(ipnet, asn, maxlen)

        for asn, ipnet, maxlen in self.server.db.get_withdrawals6(serial):
            self.withdraw_ipv6(ipnet, asn, maxlen)

        self.send_endofdata(self.serial)

    RESET_TYPE = 2

    def handle_reset(self):
        dbg(">Reset")
        self.session_id += 1
        self.server.db.set_serial(0)
        self.send_cacheresponse()

        for asn, ipnet, maxlen in self.server.db.get_announcements4(self.serial):
            self.announce_ipv4(ipnet, asn, maxlen)

        for asn, ipnet, maxlen in self.server.db.get_announcements6(self.serial):
            self.announce_ipv6(ipnet, asn, maxlen)

        self.send_endofdata(self.serial)

    ERROR_TYPE = 10

    def handle_error(self, buf: bytes):
        dbg(f">Error: {str(buf)}")
        self.server.shutdown()
        self.server.stopped = True
        raise ConnectionError("Received an RPKI error packet from FRR. Exiting")

    def handle(self):
        while True:
            b = self.request.recv(self.HEADER_LEN, socket.MSG_WAITALL)
            if len(b) == 0:
                break
            proto_ver, pdu_type, sess_id, length = self.decode_header(b)
            dbg(
                f">Header proto_ver={proto_ver} pdu_type={pdu_type} sess_id={sess_id} length={length}"
            )

            if sess_id:
                self.session_id = sess_id

            if pdu_type == self.SERIAL_QUERY_TYPE:
                b = self.request.recv(
                    self.SERIAL_QUERY_LEN - self.HEADER_LEN, socket.MSG_WAITALL
                )
                self.handle_serial_query(b, sess_id)

            elif pdu_type == self.RESET_TYPE:
                self.handle_reset()

            elif pdu_type == self.ERROR_TYPE:
                b = self.request.recv(length - self.HEADER_LEN, socket.MSG_WAITALL)
                self.handle_error(b)


class ThreadedTCPServer(socketserver.ThreadingMixIn, socketserver.TCPServer):
    def __init__(
        self, bind: Tuple[str, int], handler: Type[RTRConnHandler], db: RTRDatabase
    ) -> None:
        super().__init__(bind, handler)
        self.db = db


def main():
    db = RTRDatabase(VRPS_FILE)
    server = ThreadedTCPServer((LISTEN_HOST, LISTEN_PORT), RTRConnHandler, db)
    dbg(f"Server listening on {LISTEN_HOST} port {LISTEN_PORT}")
    server.serve_forever()


if __name__ == "__main__":
    if len(sys.argv) > 1:
        f = open(sys.argv[1], "w")
        sys.__stdout__ = f
        sys.stdout = f
        sys.__stderr__ = f
        sys.stderr = f

    main()



# -*- coding: utf-8 eval: (blacken-mode 1) -*-
# SPDX-License-Identifier: GPL-2.0-or-later
#
# July 11 2021, Christian Hopps <chopps@labn.net>
#
# Copyright (c) 2021-2023, LabN Consulting, L.L.C
#
import ipaddress
import os

from munet import cli
from munet.base import BaseMunet, LinuxNamespace


class Node(LinuxNamespace):
    """Node (mininet compat)."""

    def __init__(self, name, rundir=None, **kwargs):
        nkwargs = {}

        if "unet" in kwargs:
            nkwargs["unet"] = kwargs["unet"]
        if "private_mounts" in kwargs:
            nkwargs["private_mounts"] = kwargs["private_mounts"]
        if "logger" in kwargs:
            nkwargs["logger"] = kwargs["logger"]

        # This is expected by newer munet CLI code
        self.config_dirname = ""
        self.config = {"kind": "frr"}
        self.mgmt_ip = None
        self.mgmt_ip6 = None

        super().__init__(name, **nkwargs)

        self.rundir = self.unet.rundir.joinpath(self.name)

    def cmd(self, cmd, **kwargs):
        """Execute a command, joins stdout, stderr, ignores exit status."""

        return super(Node, self).cmd_legacy(cmd, **kwargs)

    def config_host(self, lo="up", **params):
        """Called by Micronet when topology is built (but not started)."""
        # mininet brings up loopback here.
        del params
        del lo

    def intfNames(self):
        return self.intfs

    def terminate(self):
        return

    def add_vlan(self, vlanname, linkiface, vlanid):
        self.logger.debug("Adding VLAN interface: %s (%s)", vlanname, vlanid)
        ip_path = self.get_exec_path("ip")
        assert ip_path, "XXX missing ip command!"
        self.cmd_raises(
            [
                ip_path,
                "link",
                "add",
                "link",
                linkiface,
                "name",
                vlanname,
                "type",
                "vlan",
                "id",
                vlanid,
            ]
        )
        self.cmd_raises([ip_path, "link", "set", "dev", vlanname, "up"])

    def add_loop(self, loopname):
        self.logger.debug("Adding Linux iface: %s", loopname)
        ip_path = self.get_exec_path("ip")
        assert ip_path, "XXX missing ip command!"
        self.cmd_raises([ip_path, "link", "add", loopname, "type", "dummy"])
        self.cmd_raises([ip_path, "link", "set", "dev", loopname, "up"])

    def add_l3vrf(self, vrfname, tableid):
        self.logger.debug("Adding Linux VRF: %s", vrfname)
        ip_path = self.get_exec_path("ip")
        assert ip_path, "XXX missing ip command!"
        self.cmd_raises(
            [ip_path, "link", "add", vrfname, "type", "vrf", "table", tableid]
        )
        self.cmd_raises([ip_path, "link", "set", "dev", vrfname, "up"])

    def del_iface(self, iface):
        self.logger.debug("Removing Linux Iface: %s", iface)
        ip_path = self.get_exec_path("ip")
        assert ip_path, "XXX missing ip command!"
        self.cmd_raises([ip_path, "link", "del", iface])

    def attach_iface_to_l3vrf(self, ifacename, vrfname):
        self.logger.debug("Attaching Iface %s to Linux VRF %s", ifacename, vrfname)
        ip_path = self.get_exec_path("ip")
        assert ip_path, "XXX missing ip command!"
        if vrfname:
            self.cmd_raises(
                [ip_path, "link", "set", "dev", ifacename, "master", vrfname]
            )
        else:
            self.cmd_raises([ip_path, "link", "set", "dev", ifacename, "nomaster"])

    set_cwd = LinuxNamespace.set_ns_cwd


class Topo(object):  # pylint: disable=R0205
    def __init__(self, *args, **kwargs):
        raise Exception("Remove Me")


class Mininet(BaseMunet):
    """
    Mininet using Micronet.
    """

    g_mnet_inst = None

    def __init__(self, rundir=None, pytestconfig=None, logger=None):
        """
        Create a Micronet.
        """
        if Mininet.g_mnet_inst is not None:
            Mininet.g_mnet_inst.stop()
        Mininet.g_mnet_inst = self

        self.configured_hosts = set()
        self.host_params = {}
        self.prefix_len = 8

        # SNMPd used to require this, which was set int he mininet shell
        # that all commands executed from. This is goofy default so let's not
        # do it if we don't have to. The snmpd.conf files have been updated
        # to set permissions to root:frr 770 to make this unneeded in that case
        # os.umask(0)

        super(Mininet, self).__init__(
            pid=False, rundir=rundir, pytestconfig=pytestconfig, logger=logger
        )

        # From munet/munet/native.py
        with open(os.path.join(self.rundir, "nspid"), "w", encoding="ascii") as f:
            f.write(f"{self.pid}\n")

        with open(os.path.join(self.rundir, "nspids"), "w", encoding="ascii") as f:
            f.write(f'{" ".join([str(x) for x in self.pids])}\n')

        hosts_file = os.path.join(self.rundir, "hosts.txt")
        with open(hosts_file, "w", encoding="ascii") as hf:
            hf.write(
                f"""127.0.0.1\tlocalhost {self.name}
::1\tip6-localhost ip6-loopback
fe00::0\tip6-localnet
ff00::0\tip6-mcastprefix
ff02::1\tip6-allnodes
ff02::2\tip6-allrouters
"""
            )
        self.bind_mount(hosts_file, "/etc/hosts")

        # Common CLI commands for any topology
        cdict = {
            "commands": [
                #
                # Window commands.
                #
                {
                    "name": "pcap",
                    "format": "pcap NETWORK",
                    "help": (
                        "capture packets from NETWORK into file capture-NETWORK.pcap"
                        " the command is run within a new window which also shows"
                        " packet summaries. NETWORK can also be an interface specified"
                        " as HOST:INTF. To capture inside the host namespace."
                    ),
                    "exec": "tshark -s 9200 -i {0} -P -w capture-{0}.pcap",
                    "top-level": True,
                    "new-window": {"background": True},
                },
                {
                    "name": "term",
                    "format": "term HOST [HOST ...]",
                    "help": "open terminal[s] (TMUX or XTerm) on HOST[S], * for all",
                    "exec": "bash",
                    "new-window": True,
                },
                {
                    "name": "vtysh",
                    "exec": "/usr/bin/vtysh",
                    "format": "vtysh ROUTER [ROUTER ...]",
                    "new-window": True,
                    "kinds": ["frr"],
                },
                {
                    "name": "xterm",
                    "format": "xterm HOST [HOST ...]",
                    "help": "open XTerm[s] on HOST[S], * for all",
                    "exec": "bash",
                    "new-window": {
                        "forcex": True,
                    },
                },
                {
                    "name": "logd",
                    "exec": "tail -F %RUNDIR%/{}.log",
                    "format": "logd HOST [HOST ...] DAEMON",
                    "help": (
                        "tail -f on the logfile of the given "
                        "DAEMON for the given HOST[S]"
                    ),
                    "new-window": True,
                },
                {
                    "name": "stdlog",
                    "exec": (
                        "[ -e %RUNDIR%/frr.log ] && tail -F %RUNDIR%/frr.log "
                        "|| tail -F /var/log/frr.log"
                    ),
                    "format": "stdlog HOST [HOST ...]",
                    "help": "tail -f on the `frr.log` for the given HOST[S]",
                    "new-window": True,
                },
                {
                    "name": "stdout",
                    "exec": "tail -F %RUNDIR%/{0}.err",
                    "format": "stdout HOST [HOST ...] DAEMON",
                    "help": (
                        "tail -f on the stdout of the given DAEMON for the given HOST[S]"
                    ),
                    "new-window": True,
                },
                {
                    "name": "stderr",
                    "exec": "tail -F %RUNDIR%/{0}.out",
                    "format": "stderr HOST [HOST ...] DAEMON",
                    "help": (
                        "tail -f on the stderr of the given DAEMON for the given HOST[S]"
                    ),
                    "new-window": True,
                },
                #
                # Non-window commands.
                #
                {
                    "name": "",
                    "exec": "vtysh -c '{}'",
                    "format": "[ROUTER ...] COMMAND",
                    "help": "execute vtysh COMMAND on the router[s]",
                    "kinds": ["frr"],
                },
                {
                    "name": "sh",
                    "format": "[HOST ...] sh <SHELL-COMMAND>",
                    "help": "execute <SHELL-COMMAND> on hosts",
                    "exec": "{}",
                },
                {
                    "name": "shi",
                    "format": "[HOST ...] shi <INTERACTIVE-COMMAND>",
                    "help": "execute <INTERACTIVE-COMMAND> on HOST[s]",
                    "exec": "{}",
                    "interactive": True,
                },
            ]
        }

        cli.add_cli_config(self, cdict)

        shellopt = self.cfgopt.get_option_list("--shell")
        if "all" in shellopt or "." in shellopt:
            self.run_in_window("bash", title="munet")

        # This is expected by newer munet CLI code
        self.config_dirname = ""
        self.config = {}

        self.logger.debug("%s: Creating", self)

    def __str__(self):
        return "Mininet()"

    def configure_hosts(self):
        """
        Configure hosts once the topology has been built.

        This function can be called multiple times if routers are added to the topology
        later.
        """
        if not self.hosts:
            return

        self.logger.debug("Configuring hosts: %s", self.hosts.keys())

        for name in sorted(self.hosts.keys()):
            if name in self.configured_hosts:
                continue

            host = self.hosts[name]
            first_intf = host.intfs[0] if host.intfs else None
            params = self.host_params[name]

            if first_intf and "ip" in params:
                ip = params["ip"]
                i = ip.find("/")
                if i == -1:
                    plen = self.prefix_len
                else:
                    plen = int(ip[i + 1 :])
                    ip = ip[:i]

                host.cmd_raises("ip addr add {}/{} dev {}".format(ip, plen, first_intf))

                # can be used by munet cli
                host.mgmt_ip = ipaddress.ip_address(ip)

            if "defaultRoute" in params:
                host.cmd_raises(
                    "ip route add default {}".format(params["defaultRoute"])
                )

            host.config_host()

            self.configured_hosts.add(name)

    def add_host(self, name, cls=Node, **kwargs):
        """Add a host to micronet."""

        self.host_params[name] = kwargs
        super(Mininet, self).add_host(name, cls=cls, **kwargs)

    def start(self):
        """Start the micronet topology."""
        pcapopt = self.cfgopt.get_option_list("--pcap")
        if "all" in pcapopt:
            pcapopt = self.switches.keys()
        for pcap in pcapopt:
            if ":" in pcap:
                host, intf = pcap.split(":")
                pcap = f"{host}-{intf}"
                host = self.hosts[host]
            else:
                host = self
                intf = pcap
            pcapfile = f"{self.rundir}/capture-{pcap}.pcap"
            host.run_in_window(
                f"tshark -s 9200 -i {intf} -P -w {pcapfile}",
                background=True,
                title=f"cap:{pcap}",
            )

        self.logger.debug("%s: Starting (no-op).", self)

    def stop(self):
        """Stop the mininet topology (deletes)."""
        self.logger.debug("%s: Stopping (deleting).", self)

        self.delete()

        self.logger.debug("%s: Stopped (deleted).", self)

        if Mininet.g_mnet_inst == self:
            Mininet.g_mnet_inst = None

    def cli(self):
        cli.cli(self)



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-or-later
"""
Usage: frr_babeltrace.py trace_path

FRR pushes data into lttng tracepoints in the least overhead way possible
i.e. as binary-data/crf_arrays. These traces need to be converted into pretty
strings for easy greping etc. This script is a babeltrace python plugin for
that pretty printing.

Copyright (C) 2021  NVIDIA Corporation
Anuradha Karuppiah
"""

import ipaddress
import socket
import sys

import babeltrace


########################### common parsers - start ############################
def print_ip_addr(field_val):
    """
    pretty print "struct ipaddr"
    """
    if field_val[0] == socket.AF_INET:
        addr = [str(fv) for fv in field_val[4:8]]
        return str(ipaddress.IPv4Address(".".join(addr)))

    if field_val[0] == socket.AF_INET6:
        tmp = "".join("%02x" % fb for fb in field_val[4:])
        addr = []
        while tmp:
            addr.append(tmp[:4])
            tmp = tmp[4:]
        addr = ":".join(addr)
        return str(ipaddress.IPv6Address(addr))

    if not field_val[0]:
        return ""

    return field_val


def print_mac(field_val):
    """
    pretty print "u8 mac[6]"
    """
    return ":".join("%02x" % fb for fb in field_val)


def print_net_ipv4_addr(field_val):
    """
    pretty print ctf_integer_network ipv4
    """
    return str(ipaddress.IPv4Address(field_val))


def print_esi(field_val):
    """
    pretty print ethernet segment id, esi_t
    """
    return ":".join("%02x" % fb for fb in field_val)


def get_field_list(event):
    """
    only fetch fields added via the TP, skip metadata etc.
    """
    return event.field_list_with_scope(babeltrace.CTFScope.EVENT_FIELDS)


def parse_event(event, field_parsers):
    """
    Wild card event parser; doesn't make things any prettier
    """
    field_list = get_field_list(event)
    field_info = {}
    for field in field_list:
        if field in field_parsers:
            field_parser = field_parsers.get(field)
            field_info[field] = field_parser(event.get(field))
        else:
            field_info[field] = event.get(field)
    print(event.name, field_info)


def print_family_str(field_val):
    """
    pretty print kernel family to string
    """
    if field_val == socket.AF_INET:
        cmd_str = "ipv4"
    elif field_val == socket.AF_INET6:
        cmd_str = "ipv6"
    elif field_val == socket.AF_BRIDGE:
        cmd_str = "bridge"
    elif field_val == 128:  # RTNL_FAMILY_IPMR:
        cmd_str = "ipv4MR"
    elif field_val == 129:  # RTNL_FAMILY_IP6MR:
        cmd_str = "ipv6MR"
    else:
        cmd_str = "Invalid family"

    return cmd_str


############################ common parsers - end #############################


############################ evpn parsers - start #############################
def parse_frr_bgp_evpn_mac_ip_zsend(event):
    """
    bgp evpn mac-ip parser; raw format -
    ctf_array(unsigned char, mac, &pfx->prefix.macip_addr.mac,
            sizeof(struct ethaddr))
    ctf_array(unsigned char, ip, &pfx->prefix.macip_addr.ip,
            sizeof(struct ipaddr))
    ctf_integer_network_hex(unsigned int, vtep, vtep.s_addr)
    ctf_array(unsigned char, esi, esi, sizeof(esi_t))
    """
    field_parsers = {
        "ip": print_ip_addr,
        "mac": print_mac,
        "esi": print_esi,
        "vtep": print_net_ipv4_addr,
    }

    parse_event(event, field_parsers)


def parse_frr_bgp_evpn_bum_vtep_zsend(event):
    """
    bgp evpn bum-vtep parser; raw format -
    ctf_integer_network_hex(unsigned int, vtep,
            pfx->prefix.imet_addr.ip.ipaddr_v4.s_addr)

    """
    field_parsers = {"vtep": print_net_ipv4_addr}

    parse_event(event, field_parsers)


def parse_frr_bgp_evpn_mh_nh_rmac_send(event):
    """
    bgp evpn nh-rmac parser; raw format -
    ctf_array(unsigned char, rmac, &nh->rmac, sizeof(struct ethaddr))
    """
    field_parsers = {"rmac": print_mac}

    parse_event(event, field_parsers)


def parse_frr_bgp_evpn_mh_local_es_add_zrecv(event):
    """
    bgp evpn local-es parser; raw format -
    ctf_array(unsigned char, esi, esi, sizeof(esi_t))
    ctf_integer_network_hex(unsigned int, vtep, vtep.s_addr)
    """
    field_parsers = {"esi": print_esi, "vtep": print_net_ipv4_addr}

    parse_event(event, field_parsers)


def parse_frr_bgp_evpn_mh_local_es_del_zrecv(event):
    """
    bgp evpn local-es parser; raw format -
    ctf_array(unsigned char, esi, esi, sizeof(esi_t))
    """
    field_parsers = {"esi": print_esi}

    parse_event(event, field_parsers)


def parse_frr_bgp_evpn_mh_local_es_evi_add_zrecv(event):
    """
    bgp evpn local-es-evi parser; raw format -
    ctf_array(unsigned char, esi, esi, sizeof(esi_t))
    """
    field_parsers = {"esi": print_esi}

    parse_event(event, field_parsers)


def parse_frr_bgp_evpn_mh_local_es_evi_del_zrecv(event):
    """
    bgp evpn local-es-evi parser; raw format -
    ctf_array(unsigned char, esi, esi, sizeof(esi_t))
    """
    field_parsers = {"esi": print_esi}

    parse_event(event, field_parsers)


def parse_frr_bgp_evpn_mh_es_evi_vtep_add(event):
    """
    bgp evpn remote ead evi remote vtep add; raw format -
    ctf_array(unsigned char, esi, esi, sizeof(esi_t))
    """
    field_parsers = {"esi": print_esi,
                     "vtep": print_net_ipv4_addr}

    parse_event(event, field_parsers)


def parse_frr_bgp_evpn_mh_es_evi_vtep_del(event):
    """
    bgp evpn remote ead evi remote vtep del; raw format -
    ctf_array(unsigned char, esi, esi, sizeof(esi_t))
    """
    field_parsers = {"esi": print_esi,
                     "vtep": print_net_ipv4_addr}

    parse_event(event, field_parsers)


def parse_frr_bgp_evpn_mh_local_ead_es_evi_route_upd(event):
    """
    bgp evpn local ead evi vtep; raw format -
    ctf_array(unsigned char, esi, esi, sizeof(esi_t))
    """
    field_parsers = {"esi": print_esi,
                     "vtep": print_net_ipv4_addr}

    parse_event(event, field_parsers)


def parse_frr_bgp_evpn_mh_local_ead_es_evi_route_del(event):
    """
    bgp evpn local ead evi vtep del; raw format -
    ctf_array(unsigned char, esi, esi, sizeof(esi_t))
    """
    field_parsers = {"esi": print_esi,
                     "vtep": print_net_ipv4_addr}

    parse_event(event, field_parsers)


def parse_frr_bgp_evpn_local_vni_add_zrecv(event):
    """
    bgp evpn local-vni parser; raw format -
    ctf_integer_network_hex(unsigned int, vtep, vtep.s_addr)
    ctf_integer_network_hex(unsigned int, mc_grp, mc_grp.s_addr)
    """
    field_parsers = {"vtep": print_net_ipv4_addr,
                     "mc_grp": print_net_ipv4_addr}

    parse_event(event, field_parsers)


def parse_frr_bgp_evpn_local_l3vni_add_zrecv(event):
    """
    bgp evpn local-l3vni parser; raw format -
    ctf_integer_network_hex(unsigned int, vtep, vtep.s_addr)
    ctf_array(unsigned char, svi_rmac, svi_rmac, sizeof(struct ethaddr))
    ctf_array(unsigned char, vrr_rmac, vrr_rmac, sizeof(struct ethaddr))
    """
    field_parsers = {
        "vtep": print_net_ipv4_addr,
        "svi_rmac": print_mac,
        "vrr_rmac": print_mac,
    }

    parse_event(event, field_parsers)


def parse_frr_bgp_evpn_local_macip_add_zrecv(event):
    """
    bgp evpn local-mac-ip parser; raw format -
    ctf_array(unsigned char, ip, ip, sizeof(struct ipaddr))
    ctf_array(unsigned char, mac, mac, sizeof(struct ethaddr))
    ctf_array(unsigned char, esi, esi, sizeof(esi_t))
    """
    field_parsers = {"ip": print_ip_addr,
                     "mac": print_mac,
                     "esi": print_esi}

    parse_event(event, field_parsers)


def parse_frr_bgp_evpn_local_macip_del_zrecv(event):
    """
    bgp evpn local-mac-ip del parser; raw format -
    ctf_array(unsigned char, ip, ip, sizeof(struct ipaddr))
    ctf_array(unsigned char, mac, mac, sizeof(struct ethaddr))
    """
    field_parsers = {"ip": print_ip_addr,
                     "mac": print_mac}

    parse_event(event, field_parsers)


def parse_frr_bgp_evpn_advertise_type5(event):
    """
    local originated type-5 route
    """
    field_parsers = {
        "ip": print_ip_addr,
        "rmac": print_mac,
        "vtep": print_net_ipv4_addr,
    }

    parse_event(event, field_parsers)


def parse_frr_bgp_evpn_withdraw_type5(event):
    """
    local originated type-5 route withdraw
    """
    field_parsers = {"ip": print_ip_addr}

    parse_event(event, field_parsers)


############################ evpn parsers - end *#############################


def main():
    """
    FRR lttng trace output parser; babel trace plugin
    """
    event_parsers = {"frr_bgp:evpn_mac_ip_zsend":
                     parse_frr_bgp_evpn_mac_ip_zsend,
                     "frr_bgp:evpn_bum_vtep_zsend":
                     parse_frr_bgp_evpn_bum_vtep_zsend,
                     "frr_bgp:evpn_mh_nh_rmac_zsend":
                     parse_frr_bgp_evpn_mh_nh_rmac_send,
                     "frr_bgp:evpn_mh_local_es_add_zrecv":
                     parse_frr_bgp_evpn_mh_local_es_add_zrecv,
                     "frr_bgp:evpn_mh_local_es_del_zrecv":
                     parse_frr_bgp_evpn_mh_local_es_del_zrecv,
                     "frr_bgp:evpn_mh_local_es_evi_add_zrecv":
                     parse_frr_bgp_evpn_mh_local_es_evi_add_zrecv,
                     "frr_bgp:evpn_mh_local_es_evi_del_zrecv":
                     parse_frr_bgp_evpn_mh_local_es_evi_del_zrecv,
                     "frr_bgp:evpn_mh_es_evi_vtep_add":
                     parse_frr_bgp_evpn_mh_es_evi_vtep_add,
                     "frr_bgp:evpn_mh_es_evi_vtep_del":
                     parse_frr_bgp_evpn_mh_es_evi_vtep_del,
                     "frr_bgp:evpn_mh_local_ead_es_evi_route_upd":
                     parse_frr_bgp_evpn_mh_local_ead_es_evi_route_upd,
                     "frr_bgp:evpn_mh_local_ead_es_evi_route_del":
                     parse_frr_bgp_evpn_mh_local_ead_es_evi_route_del,
                     "frr_bgp:evpn_local_vni_add_zrecv":
                     parse_frr_bgp_evpn_local_vni_add_zrecv,
                     "frr_bgp:evpn_local_l3vni_add_zrecv":
                     parse_frr_bgp_evpn_local_l3vni_add_zrecv,
                     "frr_bgp:evpn_local_macip_add_zrecv":
                     parse_frr_bgp_evpn_local_macip_add_zrecv,
                     "frr_bgp:evpn_local_macip_del_zrecv":
                     parse_frr_bgp_evpn_local_macip_del_zrecv,
                     "frr_bgp:evpn_advertise_type5":
                     parse_frr_bgp_evpn_advertise_type5,
                     "frr_bgp:evpn_withdraw_type5":
                     parse_frr_bgp_evpn_withdraw_type5,
}

    # get the trace path from the first command line argument
    trace_path = sys.argv[1]

    # grab events
    trace_collection = babeltrace.TraceCollection()
    trace_collection.add_traces_recursive(trace_path, "ctf")

    for event in trace_collection.events:
        if event.name in event_parsers:
            event_parser = event_parsers.get(event.name)
            event_parser(event)
        else:
            parse_event(event, {})


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# SPDX-License-Identifier: ISC
#
# mcast-rx.py
#
# Copyright (c) 2018 Cumulus Networks, Inc.
#
"""
Subscribe to a multicast group so that the kernel sends an IGMP JOIN
for the multicast group we subscribed to.
"""

import argparse
import logging
import re
import os
import socket
import subprocess
import struct
import sys
import time


def ifname_to_ifindex(ifname):
    output = subprocess.check_output(
        "ip link show %s" % ifname, shell=True, universal_newlines=True
    )
    first_line = output.split("\n")[0]
    re_index = re.search("^(\d+):", first_line)

    if re_index:
        return int(re_index.group(1))

    log.error("Could not parse the ifindex for %s out of\n%s" % (ifname, first_line))
    return None


# Thou shalt be root
if os.geteuid() != 0:
    sys.stderr.write("ERROR: You must have root privileges\n")
    sys.exit(1)


logging.basicConfig(
    level=logging.DEBUG, format="%(asctime)s %(levelname)5s: %(message)s"
)

# Color the errors and warnings in red
logging.addLevelName(
    logging.ERROR, "\033[91m  %s\033[0m" % logging.getLevelName(logging.ERROR)
)
logging.addLevelName(
    logging.WARNING, "\033[91m%s\033[0m" % logging.getLevelName(logging.WARNING)
)
log = logging.getLogger(__name__)

parser = argparse.ArgumentParser(description="Multicast RX utility")

parser.add_argument("group", help="Multicast IP")
parser.add_argument("ifname", help="Interface name")
parser.add_argument("--port", help="UDP port", default=1000)
parser.add_argument("--sleep", help="Time to sleep before we stop waiting", default=5)
args = parser.parse_args()

# Create the datagram socket
sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
sock.bind((args.group, args.port))

newpid = os.fork()

if newpid == 0:
    ifindex = ifname_to_ifindex(args.ifname)
    mreq = struct.pack(
        "=4sLL", socket.inet_aton(args.group), socket.INADDR_ANY, ifindex
    )
    sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
    time.sleep(float(args.sleep))
    sock.close()



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

# Copyright (c) 2024 by
# Donatas Abraitis <donatas@opensourcerouting.org>
#

import os
import sys
import json
import pytest
import functools

CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
from lib import topotest
from lib.topogen import Topogen, get_topogen

pytestmark = [pytest.mark.bgpd]


def setup_module(mod):
    topodef = {"s1": ("r1", "r2", "r3")}
    tgen = Topogen(topodef, mod.__name__)
    tgen.start_topology()

    router_list = tgen.routers()

    for _, (rname, router) in enumerate(router_list.items(), 1):
        router.load_frr_config(os.path.join(CWD, "{}/frr.conf".format(rname)))

    tgen.start_router()


def teardown_module(mod):
    tgen = get_topogen()
    tgen.stop_topology()


def test_bgp_self_prefix():
    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    r1 = tgen.gears["r1"]
    r3 = tgen.gears["r3"]

    def _bgp_converge():
        output = json.loads(r1.vtysh_cmd("show bgp ipv4 unicast json"))
        expected = {
            "routes": {
                "10.0.0.2/32": [
                    {
                        "valid": True,
                        "path": "",
                        "nexthops": [
                            {"ip": "10.0.0.2", "hostname": "r2", "afi": "ipv4"}
                        ],
                    }
                ],
                "10.0.0.3/32": [
                    {
                        "valid": True,
                        "path": "65003",
                        "nexthops": [
                            {"ip": "10.0.0.3", "hostname": "r3", "afi": "ipv4"}
                        ],
                    }
                ],
            }
        }

        return topotest.json_cmp(output, expected)

    test_func = functools.partial(
        _bgp_converge,
    )
    _, result = topotest.run_and_expect(test_func, None, count=30, wait=1)
    assert result is None, "Can't converge"

    def _bgp_check_received_routes():
        output = json.loads(r3.vtysh_cmd("show bgp ipv4 unicast json"))
        expected = {
            "routes": {
                "10.0.0.2/32": [
                    {
                        "valid": True,
                        "bestpath": True,
                        "nexthops": [
                            {"ip": "10.0.0.1", "hostname": "r1", "afi": "ipv4"}
                        ],
                    }
                ],
            }
        }

        return topotest.json_cmp(output, expected)

    test_func = functools.partial(
        _bgp_check_received_routes,
    )
    _, result = topotest.run_and_expect(test_func, None, count=30, wait=1)
    assert result is None, "Can't see 10.0.0.2/32"


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))
